import os

from matio import load_from_mat
from matio.mat_opaque_tools import MatioOpaque


def test_handle_load_v7():
    file_path_v7 = os.path.join(os.path.dirname(__file__), "test_handle_v7.mat")
    prop_names = ["Name", "Next"]

    matdict = load_from_mat(file_path_v7, raw_data=False)

    assert isinstance(matdict["obj1"], MatioOpaque)
    assert isinstance(matdict["obj2"], MatioOpaque)

    assert matdict["obj1"].classname == "Node"
    assert matdict["obj2"].classname == "Node"

    assert set(matdict["obj1"].properties.keys()) == set(prop_names)
    assert set(matdict["obj2"].properties.keys()) == set(prop_names)

    assert matdict["obj1"].properties["Next"] is matdict["obj2"]
    assert matdict["obj2"].properties["Next"] is matdict["obj1"]


def test_handle_load_v73():
    file_path_v73 = os.path.join(os.path.dirname(__file__), "test_handle_v73.mat")
    prop_names = ["Name", "Next"]

    matdict = load_from_mat(file_path_v73, raw_data=False)

    assert isinstance(matdict["obj1"], MatioOpaque)
    assert isinstance(matdict["obj2"], MatioOpaque)

    assert matdict["obj1"].classname == "Node"
    assert matdict["obj2"].classname == "Node"

    assert set(matdict["obj1"].properties.keys()) == set(prop_names)
    assert set(matdict["obj2"].properties.keys()) == set(prop_names)

    assert matdict["obj1"].properties["Next"] is matdict["obj2"]
    assert matdict["obj2"].properties["Next"] is matdict["obj1"]



import os

import numpy as np
import pytest

from matio import load_from_mat

params = [
    (
        np.array([5], dtype="timedelta64[s]").reshape(1, 1),
        "dur1",
    ),
    (
        np.array([5], dtype="timedelta64[m]").reshape(1, 1),
        "dur2",
    ),
    (
        np.array([5], dtype="timedelta64[h]").reshape(1, 1),
        "dur3",
    ),
    (
        np.array([5], dtype="timedelta64[D]").reshape(1, 1),
        "dur4",
    ),
    (
        np.array([1, 2, 3], dtype="timedelta64[Y]").reshape(1, 3),
        "dur8",
    ),
    (
        (np.timedelta64(1, "h") + np.timedelta64(2, "m") + np.timedelta64(3, "s"))
        .astype("timedelta64[ms]")
        .reshape(1, 1),
        "dur5",
    ),
    (
        np.array([10, 20, 30, 40, 50, 60], dtype="timedelta64[s]").reshape(2, 3),
        "dur6",
    ),
    (
        np.array([], dtype="datetime64[ms]"),
        "dur7",
    ),
]


@pytest.mark.parametrize(
    "expected_array, var_name",
    params,
    ids=[
        "duration-seconds-v7",
        "duration-minutes-v7",
        "duration-hours-v7",
        "duration-days-v7",
        "duration-years-v7",
        "duration-base-v7",
        "duration-array-v7",
        "duration-empty-v7",
    ],
)
def test_duration_load_v7(expected_array, var_name):
    file_path_v7 = os.path.join(os.path.dirname(__file__), "test_duration_v7.mat")
    matdict = load_from_mat(file_path_v7, raw_data=False)

    assert var_name in matdict
    np.testing.assert_array_equal(matdict[var_name], expected_array)


@pytest.mark.parametrize(
    "expected_array, var_name",
    params,
    ids=[
        "duration-seconds-v7.3",
        "duration-minutes-v7.3",
        "duration-hours-v7.3",
        "duration-days-v7.3",
        "duration-years-v7.3",
        "duration-base-v7.3",
        "duration-array-v7.3",
        "duration-empty-v7.3",
    ],
)
def test_duration_load_v73(expected_array, var_name):
    file_path_v73 = os.path.join(os.path.dirname(__file__), "test_duration_v73.mat")
    matdict = load_from_mat(file_path_v73, raw_data=False)

    assert var_name in matdict
    np.testing.assert_array_equal(matdict[var_name], expected_array)



import os

import numpy as np
import pandas as pd
import pytest

from matio import load_from_mat


def make_cell_entry(sub_arr):
    arr = np.empty(3, dtype=object)
    arr[:] = [sub_arr, sub_arr, sub_arr]
    return arr


params_base = [
    (
        pd.DataFrame({"Var1": [1.1, 2.2, 3.3], "Var2": [4.4, 5.5, 6.6]}),
        "T1",
    ),
    (
        pd.DataFrame(
            {
                "Var1": ["apple", "banana", "cherry"],
            }
        ),
        "T2",
    ),
    (
        pd.DataFrame(
            {
                "Time": np.array(
                    [
                        "2020-01-01T00:00:00.000",
                        "2020-01-02T00:00:00.000",
                        "2020-01-03T00:00:00.000",
                    ],
                    dtype="datetime64[ms]",
                ),
                "Duration": np.array([30, 60, 90], dtype="timedelta64[s]"),
            }
        ),
        "T3",
    ),
    (
        pd.DataFrame(
            {
                "Var1": [
                    np.array([[1.0]]),
                    np.array(["text"]),
                    np.array([["2023-01-01T00:00:00.000"]], dtype="datetime64[ms]"),
                ],
            }
        ),
        "T5",
    ),
    (
        pd.DataFrame(
            {
                "Var1": [1.1, np.nan, 3.3],
                "Var2": np.array(["A", "", "C"]),
            }
        ),
        "T6",
    ),
    (
        pd.DataFrame(
            {
                "X": np.array([10.0, 20.0, 30.0]),
                "Y": np.array([100.0, 200.0, 300.0]),
            }
        ),
        "T7",
    ),
    (
        pd.DataFrame(
            {
                "Col1": np.array([1.0, 2.0, 3.0]),
                "Col2": np.array([4.0, 5.0, 6.0]),
            },
            index=["R1", "R2", "R3"],
        ),
        "T8",
    ),
    (
        pd.DataFrame(
            {
                "Var1": np.array(
                    [
                        "2023-01-01T00:00:00.000",
                        "2023-01-02T00:00:00.000",
                        "2023-01-03T00:00:00.000",
                    ],
                    dtype="datetime64[ms]",
                ),
                "data_1": np.array([1.0, 2.0, 3.0]),
                "data_2": np.array([4.0, 5.0, 6.0]),
            }
        ),
        "T10",
    ),
]

params_obj = [
    (
        pd.DataFrame(
            {
                "C": make_cell_entry(
                    np.array(
                        [[(np.array([[123.0]]), np.array(["abc"]))]],
                        dtype=[("field1", "O"), ("field2", "O")],
                    )
                ),
                "Var2": make_cell_entry(
                    {
                        "Value": np.array([[42.0]]),
                    }
                ),
            }
        ),
        "T4",
    )
]

params_attrs = [
    (
        pd.DataFrame(
            {
                "ID": np.array([1.0, 2.0]),
                "Label": np.array(["one", "two"]),
            }
        ),
        "T9",
    ),
]


@pytest.mark.parametrize(
    "expected_df, var_name",
    params_base,
    ids=[
        "simple-table-v7",
        "table-with-strings-v7",
        "table-with-datetime-v7",
        "table-with-mixed-types-v7",
        "table-with-nan-v7",
        "table-with-row-names-v7",
        "table-with-row-names-v7",
        "table-with-multicolumn-var-v7",
    ],
)
def test_table_load_v7(expected_df, var_name):
    file_path_v7 = os.path.join(os.path.dirname(__file__), "test_table_v7.mat")
    matdict = load_from_mat(file_path_v7, raw_data=False)

    assert var_name in matdict
    pd.testing.assert_frame_equal(matdict[var_name], expected_df)


@pytest.mark.parametrize(
    "expected_df, var_name",
    params_base,
    ids=[
        "simple-table-v7.3",
        "table-with-strings-v7.3",
        "table-with-datetime-v7.3",
        "table-with-mixed-types-v7.3",
        "table-with-nan-v7.3",
        "table-with-row-names-v7.3",
        "table-with-row-names-v7.3",
        "table-with-multicolumn-var-v7.3",
    ],
)
def test_table_load_v73(expected_df, var_name):
    file_path_v73 = os.path.join(os.path.dirname(__file__), "test_table_v73.mat")
    matdict = load_from_mat(file_path_v73, raw_data=False)

    assert var_name in matdict
    pd.testing.assert_frame_equal(matdict[var_name], expected_df)


@pytest.mark.parametrize(
    "expected_df, var_name",
    params_obj,
    ids=["table-with-struct-and-object-v7"],
)
def test_table_struct_and_object_load_v7(expected_df, var_name):
    file_path_v7 = os.path.join(os.path.dirname(__file__), "test_table_v7.mat")
    matdict = load_from_mat(file_path_v7, raw_data=False)
    assert var_name in matdict
    pd.testing.assert_frame_equal(matdict[var_name], expected_df)


@pytest.mark.parametrize(
    "expected_df, var_name",
    params_obj,
    ids=["table-with-struct-and-object-v7.3"],
)
def test_table_struct_and_object_load_v73(expected_df, var_name):
    file_path_v73 = os.path.join(os.path.dirname(__file__), "test_table_v73.mat")
    matdict = load_from_mat(file_path_v73, raw_data=False)
    assert var_name in matdict
    pd.testing.assert_frame_equal(matdict[var_name], expected_df)


@pytest.mark.parametrize(
    "expected_df, var_name",
    params_attrs,
    ids=["table-with-attrs-v7"],
)
def test_table_with_attrs_load_v7(expected_df, var_name):
    file_path_v7 = os.path.join(os.path.dirname(__file__), "test_table_v7.mat")
    matdict = load_from_mat(file_path_v7, raw_data=False, add_table_attrs=True)
    expected_df.attrs = {
        "Description": "Test table with full metadata",
        "DimensionNames": ["RowId", "Features"],
        "VariableUnits": ["", "category"],
        "VariableDescriptions": ["ID number", "Category label"],
        "VariableContinuity": ["continuous", "step"],
        "UserData": np.array(
            [[(np.array(["UnitTest"]), np.array([[1.0]]))]],
            dtype=[("CreatedBy", "O"), ("Version", "O")],
        ),
    }

    assert var_name in matdict
    pd.testing.assert_frame_equal(matdict[var_name], expected_df)

    # Check attributes
    for key, value in expected_df.attrs.items():
        assert key in matdict[var_name].attrs
        if isinstance(value, np.ndarray):
            np.testing.assert_array_equal(matdict[var_name].attrs[key], value)
        else:
            assert matdict[var_name].attrs[key] == value


@pytest.mark.parametrize(
    "expected_df, var_name",
    params_attrs,
    ids=["table-with-attrs-v7.3"],
)
def test_table_with_attrs_load_v73(expected_df, var_name):
    file_path_v73 = os.path.join(os.path.dirname(__file__), "test_table_v73.mat")
    matdict = load_from_mat(file_path_v73, raw_data=False, add_table_attrs=True)
    expected_df.attrs = {
        "Description": "Test table with full metadata",
        "DimensionNames": ["RowId", "Features"],
        "VariableUnits": ["", "category"],
        "VariableDescriptions": ["ID number", "Category label"],
        "VariableContinuity": ["continuous", "step"],
        "UserData": np.array(
            [[(np.array(["UnitTest"]), np.array([[1.0]]))]],
            dtype=[("CreatedBy", "O"), ("Version", "O")],
        ),
    }

    assert var_name in matdict
    pd.testing.assert_frame_equal(matdict[var_name], expected_df)

    # Check attributes
    for key, value in expected_df.attrs.items():
        assert key in matdict[var_name].attrs
        if isinstance(value, np.ndarray):
            np.testing.assert_array_equal(matdict[var_name].attrs[key], value)
        else:
            assert matdict[var_name].attrs[key] == value



import os

import numpy as np
import pandas as pd
import pytest

from matio import load_from_mat

params_base = [
    (
        pd.DataFrame(
            {"data1": [1.0, 2.0, 3.0]},
            index=pd.Index(
                np.array(
                    ["2023-01-01", "2023-01-02", "2023-01-03"],
                    dtype="datetime64[ms]",
                ),
                name="Time",
            ),
        ),
        "tt1",
    ),
    (
        pd.DataFrame(
            {
                "data2_1": [1.0, 2.0, 3.0],
                "data2_2": [4.0, 5.0, 6.0],
            },
            index=pd.Index(
                np.array([10, 20, 30], dtype="timedelta64[s]"),
                name="Time",
            ),
        ),
        "tt2",
    ),
    (
        pd.DataFrame(
            {
                "data1": [1.0, 2.0, 3.0],
                "data3": [7.0, 8.0, 9.0],
            },
            index=pd.Index(
                np.array(
                    ["2023-01-01", "2023-01-02", "2023-01-03"],
                    dtype="datetime64[ms]",
                ),
                name="Time",
            ),
        ),
        "tt3",
    ),
    (
        pd.DataFrame(
            {
                "data1": [1.0, 2.0, 3.0],
            },
            index=pd.Index(
                np.array([0, int(1e5), int(2e5)], dtype="timedelta64[ns]"),
                name="Time",
            ),
        ),
        "tt4",
    ),
    (
        pd.DataFrame(
            {
                "data1": [1.0, 2.0, 3.0],
            },
            index=pd.Index(
                np.array([0, int(1e9), int(2e9)], dtype="timedelta64[ns]"),
                name="Time",
            ),
        ),
        "tt5",
    ),
    (
        pd.DataFrame(
            {
                "data1": [1.0, 2.0, 3.0],
            },
            index=pd.Index(
                np.array([int(10e9), int(11e9), int(12e9)], dtype="timedelta64[ns]"),
                name="Time",
            ),
        ),
        "tt6",
    ),
    (
        pd.DataFrame(
            {
                "data1": [1.0, 2.0, 3.0],
            },
            index=pd.Index(
                np.array(
                    [
                        "2020-01-01T00:00:00",
                        "2020-01-01T00:00:01",
                        "2020-01-01T00:00:02",
                    ],
                    dtype="datetime64[ns]",
                ),
                name="Time",
            ),
        ),
        "tt7",
    ),
    (
        pd.DataFrame(
            {"Pressure": [1.0, 2.0, 3.0]},
            index=pd.Index(
                np.array(
                    ["2023-01-01", "2023-01-02", "2023-01-03"],
                    dtype="datetime64[ms]",
                ),
                name="Time",
            ),
        ),
        "tt8",
    ),
    (
        pd.DataFrame(
            {
                "data1": [1.0, 2.0, 3.0],
            },
            index=pd.Index(
                np.array(
                    [
                        "2020-01-01T00:00:00",
                        "2020-04-01T00:00:01",
                        "2020-07-01T00:00:02",
                    ],
                    dtype="datetime64[M]",
                ),
                name="Time",
            ),
        ),
        "tt10",
    ),
]

param_attrs = [
    (
        pd.DataFrame(
            {"data1": [1.0, 2.0, 3.0]},
            index=pd.Index(
                np.array(
                    ["2023-01-01", "2023-01-02", "2023-01-03"],
                    dtype="datetime64[ms]",
                ),
                name="Date",
            ),
        ),
        "tt9",
    ),
]


@pytest.mark.parametrize(
    "expected_df, var_name",
    params_base,
    ids=[
        "simple-timetable-datetime-v7",
        "simple-timetable-datetime-multicolumn-v7",
        "simple-timetable-datetime-multivars-v7",
        "timetable-samplerate-v7",
        "timetable-timestep-v7",
        "timetable-samplerate-starttime-v7",
        "timetable-timestep-starttime-v7",
        "timetable-varnames-v7",
        "timetable-calendarDuration-timestep-v7",
    ],
)
def test_timetable_load_v7(expected_df, var_name):
    file_path = os.path.join(os.path.dirname(__file__), "test_timetable_v7.mat")
    matdict = load_from_mat(file_path, raw_data=False)

    assert var_name in matdict
    pd.testing.assert_frame_equal(matdict[var_name], expected_df)


@pytest.mark.parametrize(
    "expected_df, var_name",
    params_base,
    ids=[
        "simple-timetable-datetime-v7.3",
        "simple-timetable-datetime-multicolumn-v7.3",
        "simple-timetable-datetime-multivars-v7.3",
        "timetable-samplerate-v7.3",
        "timetable-timestep-v7.3",
        "timetable-samplerate-starttime-v7.3",
        "timetable-timestep-starttime-v7.3",
        "timetable-varnames-v7.3",
        "timetable-calendarDuration-timestep-v7.3",
    ],
)
def test_timetable_load_v73(expected_df, var_name):
    file_path = os.path.join(os.path.dirname(__file__), "test_timetable_v73.mat")
    matdict = load_from_mat(file_path, raw_data=False)

    assert var_name in matdict
    pd.testing.assert_frame_equal(matdict[var_name], expected_df)


@pytest.mark.parametrize(
    "expected_df, var_name",
    param_attrs,
    ids=["timetable-with-attrs-v7"],
)
def test_timetable_with_attrs_load_v7(expected_df, var_name):
    file_path = os.path.join(os.path.dirname(__file__), "test_timetable_v7.mat")
    matdict = load_from_mat(file_path, raw_data=False, add_table_attrs=True)
    expected_df.attrs = {
        "Description": "Random Description",
        "varUnits": ["m/s"],
        "varDescriptions": ["myVar"],
        "varContinuity": ["continuous"],
        "UserData": np.empty((0, 0), dtype=float),
    }

    assert var_name in matdict
    pd.testing.assert_frame_equal(matdict[var_name], expected_df)

    # Check attributes
    for key, value in expected_df.attrs.items():
        assert key in matdict[var_name].attrs
        if isinstance(value, np.ndarray):
            np.testing.assert_array_equal(matdict[var_name].attrs[key], value)
        else:
            assert matdict[var_name].attrs[key] == value


@pytest.mark.parametrize(
    "expected_df, var_name",
    param_attrs,
    ids=["timetable-with-attrs-v7.3"],
)
def test_timetable_with_attrs_load_v73(expected_df, var_name):
    file_path = os.path.join(os.path.dirname(__file__), "test_timetable_v73.mat")
    matdict = load_from_mat(file_path, raw_data=False, add_table_attrs=True)
    expected_df.attrs = {
        "Description": "Random Description",
        "varUnits": ["m/s"],
        "varDescriptions": ["myVar"],
        "varContinuity": ["continuous"],
        "UserData": np.empty((0, 0), dtype=float),
    }

    assert var_name in matdict
    pd.testing.assert_frame_equal(matdict[var_name], expected_df)

    # Check attributes
    for key, value in expected_df.attrs.items():
        assert key in matdict[var_name].attrs
        if isinstance(value, np.ndarray):
            np.testing.assert_array_equal(matdict[var_name].attrs[key], value)
        else:
            assert matdict[var_name].attrs[key] == value



import os

import numpy as np
import pytest

from matio import load_from_mat

params = [
    (
        (
            np.empty((0, 0), dtype="timedelta64[M]"),
            np.empty((0, 0), dtype="timedelta64[D]"),
            np.empty((0, 0), dtype="timedelta64[ms]"),
        ),
        "cdur1",
    ),
    (
        (
            np.array([[0]], dtype="timedelta64[M]"),
            np.array([[1, 2, 3]], dtype="timedelta64[D]"),
            np.array([[0]], dtype="timedelta64[ms]"),
        ),
        "cdur2",
    ),
    (
        (
            np.array([[0]], dtype="timedelta64[M]"),
            np.array([[7, 14]], dtype="timedelta64[D]"),
            np.array([[0]], dtype="timedelta64[ms]"),
        ),
        "cdur3",
    ),
    (
        (
            np.array([[1, 0]], dtype="timedelta64[M]"),
            np.array([[1, 2]], dtype="timedelta64[D]"),
            np.array([[0]], dtype="timedelta64[ms]"),
        ),
        "cdur4",
    ),
    (
        (
            np.array([[12, 18]], dtype="timedelta64[M]"),
            np.array([[0]], dtype="timedelta64[D]"),
            np.array([[0]], dtype="timedelta64[ms]"),
        ),
        "cdur5",
    ),
    (
        (
            np.array([[3]], dtype="timedelta64[M]"),
            np.array([[15]], dtype="timedelta64[D]"),
            np.array([[0]], dtype="timedelta64[ms]"),
        ),
        "cdur6",
    ),
    (
        (
            np.array([[1, 0], [2, 0]], dtype="timedelta64[M]"),
            np.array([[0, 5], [0, 10]], dtype="timedelta64[D]"),
            np.array([[0, 0], [0, 0]], dtype="timedelta64[ms]"),
        ),
        "cdur7",
    ),
    (
        (
            np.array([[0]], dtype="timedelta64[M]"),
            np.array([[1]], dtype="timedelta64[D]"),
            np.array([[3723000]], dtype="timedelta64[ms]"),
        ),
        "cdur8",
    ),
]


@pytest.mark.parametrize(
    "expected_array, var_name",
    params,
    ids=[
        "calendarDuration-empty-v7",
        "calendarDuration-days-v7",
        "calendarDuration-weeks-v7",
        "calendarDuration-mixed-v7",
        "calendarDuration-years-v7",
        "calendarDuration-quarters-v7",
        "calendarDuration-2D-v7",
        "calendarDuration-duration-v7",
    ],
)
def test_calendarduration_load_v7(expected_array, var_name):
    file_path_v7 = os.path.join(os.path.dirname(__file__), "test_caldur_v7.mat")
    matdict = load_from_mat(file_path_v7, raw_data=False)

    assert var_name in matdict
    assert np.array_equal(matdict[var_name][0, 0]["months"], expected_array[0])
    assert np.array_equal(matdict[var_name][0, 0]["days"], expected_array[1])
    assert np.array_equal(matdict[var_name][0, 0]["millis"], expected_array[2])


@pytest.mark.parametrize(
    "expected_array, var_name",
    params,
    ids=[
        "calendarDuration-empty-v7.3",
        "calendarDuration-days-v7.3",
        "calendarDuration-weeks-v7.3",
        "calendarDuration-mixed-v7.3",
        "calendarDuration-years-v7.3",
        "calendarDuration-quarters-v7.3",
        "calendarDuration-2D-v7.3",
        "calendarDuration-duration-v7.3",
    ],
)
def test_calendarduration_load_v73(expected_array, var_name):
    file_path_v73 = os.path.join(os.path.dirname(__file__), "test_caldur_v73.mat")
    matdict = load_from_mat(file_path_v73, raw_data=False)

    assert var_name in matdict
    assert np.array_equal(matdict[var_name][0, 0]["months"], expected_array[0])
    assert np.array_equal(matdict[var_name][0, 0]["days"], expected_array[1])
    assert np.array_equal(matdict[var_name][0, 0]["millis"], expected_array[2])



import os

import numpy as np
import pytest

from matio import load_from_mat

params = [
    (
        np.array(["blue", "green", "red"]),
        np.array([[2, 1, 0, 2]]),
        False,
        "cat1",
    ),
    (
        np.array(["high", "low", "medium"]),
        np.array([[1, 2], [0, 1]]),
        False,
        "cat2",
    ),
    (
        np.array(["cold", "warm", "hot"]),
        np.array([[0, 2, 1]]),
        False,
        "cat3",
    ),
    (
        np.array(["small", "medium", "large"]),
        np.array([[0, 1, 2]]),
        True,
        "cat4",
    ),
    (
        np.array(["low", "medium", "high"]),
        np.array([[0, 1, 2, 1, 0]]),
        False,
        "cat5",
    ),
    (
        [],
        np.empty((0, 0), dtype=np.uint8),
        False,
        "cat6",
    ),
    (
        np.array(["cat", "dog", "mouse"]),
        np.array([[0, -1, 1, 2]]),
        False,
        "cat7",
    ),
    (
        np.array(["autumn", "spring", "summer", "winter"]),
        np.array([[1, 2, 0, 3]]),
        False,
        "cat8",
    ),
    (
        np.array(["OFF", "ON", "On", "off", "on"]),
        np.array([[2, 3, 0, 1, 4]]),
        False,
        "cat9",
    ),
    (
        np.array(["maybe", "no", "yes"]),
        np.tile(np.array([[2, 2], [1, 1], [0, 0]], dtype=np.uint8), (2, 1, 1)),
        False,
        "cat10",
    ),
]


@pytest.mark.parametrize(
    "categories, codes, ordered, var_name",
    params,
    ids=[
        "categorical-basic-v7",
        "categorical-2D-v7",
        "categorical-explicit-cats-v7",
        "categorical-ordered-v7",
        "categorical-numeric-labels-v7",
        "categorical-empty-v7",
        "categorical-missing-labels-v7",
        "categorical-from-string-v7",
        "categorical-mixed-case-cats-v7",
        "categorical-3D-v7",
    ],
)
def test_categorical_load_v7(categories, codes, ordered, var_name):
    file_path_v7 = os.path.join(os.path.dirname(__file__), "test_categorical_v7.mat")
    matdict = load_from_mat(file_path_v7, raw_data=False)
    assert var_name in matdict
    assert np.array_equal(matdict[var_name].codes, codes)
    assert np.array_equal(matdict[var_name].categories, categories)
    assert np.array_equal(matdict[var_name].ordered, ordered)


@pytest.mark.parametrize(
    "categories, codes, ordered, var_name",
    params,
    ids=[
        "categorical-basic-v7.3",
        "categorical-2D-v7.3",
        "categorical-explicit-cats-v7.3",
        "categorical-ordered-v7.3",
        "categorical-numeric-labels-v7.3",
        "categorical-empty-v7.3",
        "categorical-missing-labels-v7.3",
        "categorical-from-string-v7.3",
        "categorical-mixed-case-cats-v7.3",
        "categorical-3D-v7.3",
    ],
)
def test_categorical_load_v73(categories, codes, ordered, var_name):
    file_path_v73 = os.path.join(os.path.dirname(__file__), "test_categorical_v73.mat")
    matdict = load_from_mat(file_path_v73, raw_data=False)
    assert var_name in matdict
    assert np.array_equal(matdict[var_name].codes, codes)
    assert np.array_equal(matdict[var_name].categories, categories)
    assert np.array_equal(matdict[var_name].ordered, ordered)



import os

import numpy as np
import pytest

from matio import load_from_mat

params = [
    (
        (
            np.array([[1.0, 2.0, 3.0]]).reshape(-1, 1),
            np.array(["apple", "banana", "cherry"]).reshape(-1, 1),
        ),
        "dict1",
    ),
    (
        (
            np.array(["x", "y", "z"]).reshape(-1, 1),
            np.array([10.0, 20.0, 30.0]).reshape(-1, 1),
        ),
        "dict2",
    ),
    (
        (
            np.array([["name", "age"]]).reshape(-1, 1),
            np.array([np.array(["Alice"]), np.array([25.0])], dtype=object).reshape(
                -1, 1
            ),
        ),
        "dict3",
    ),
    (
        (
            np.array([[1, 2, 3]]).reshape(-1, 1),
            np.array(["one", "two", "three"]).reshape(-1, 1),
        ),
        "dict4",
    ),
]


@pytest.mark.parametrize(
    "expected_array, var_name",
    params,
    ids=[
        "dict-numeric-key-v7",
        "dict-char-key-v7",
        "dict-mixed-val-v7",
        "dict-cell-key-v7",
    ],
)
def test_dict_load_v7(expected_array, var_name):
    file_path_v7 = os.path.join(os.path.dirname(__file__), "test_dict_v7.mat")
    matdict = load_from_mat(file_path_v7, raw_data=False)

    assert var_name in matdict
    assert np.array_equal(matdict[var_name][0], expected_array[0])
    assert np.array_equal(matdict[var_name][1], expected_array[1])


@pytest.mark.parametrize(
    "expected_array, var_name",
    params,
    ids=[
        "dict-numeric-key-v7.3",
        "dict-char-key-v7.3",
        "dict-mixed-val-v7.3",
        "dict-cell-key-v7.3",
    ],
)
def test_containermap_load_v73(expected_array, var_name):
    file_path_v73 = os.path.join(os.path.dirname(__file__), "test_dict_v73.mat")
    matdict = load_from_mat(file_path_v73, raw_data=False)

    assert var_name in matdict
    assert np.array_equal(matdict[var_name][0], expected_array[0])
    assert np.array_equal(matdict[var_name][1], expected_array[1])



"""Python library for performing I/O (currently read-only) operations on MATLAB MAT-files."""

from .mat_opaque_tools import MatioOpaque
from .matio_base import load_from_mat, save_to_mat
from .subsystem import get_matio_context

__all__ = ["load_from_mat", "save_to_mat", "get_matio_context", "MatioOpaque"]



import os

import numpy as np
import pytest
from scipy.io import loadmat

from matio import load_from_mat

sc_v7 = os.path.join(os.path.dirname(__file__), "struct_cell_v7.mat")
sc_v73 = os.path.join(os.path.dirname(__file__), "struct_cell_v73.mat")

v7_data = loadmat(sc_v7, mat_dtype=True)
variables = [(key, v7_data[key]) for key in v7_data.keys() if not key.startswith("__")]


def assert_mat_arrays_equal(actual, expected, name="var"):
    """
    Recursively assert equality between two MATLAB-like arrays (loaded as numpy arrays),
    handling nested object arrays (e.g., cells) and structured arrays (e.g., structs).
    """
    assert isinstance(actual, np.ndarray), f"{name}: actual is not a numpy array"
    assert isinstance(expected, np.ndarray), f"{name}: expected is not a numpy array"
    assert (
        actual.shape == expected.shape
    ), f"{name}: shape mismatch {actual.shape} != {expected.shape}"

    # Case 1: Structured array (MATLAB struct)
    if actual.dtype.names is not None and expected.dtype.names is not None:
        assert (
            actual.dtype.names == expected.dtype.names
        ), f"{name}: struct fields mismatch {actual.dtype.names} != {expected.dtype.names}"
        for field in actual.dtype.names:
            for idx in np.ndindex(actual.shape):
                subname = f"{name}[{idx}].{field}"
                act_val = actual[idx][field]
                exp_val = expected[idx][field]
                assert_mat_arrays_equal(
                    np.asarray(act_val), np.asarray(exp_val), name=subname
                )

    # Case 2: Object array (e.g., cells or mixed struct/cell)
    elif actual.dtype == object or expected.dtype == object:
        for idx in np.ndindex(actual.shape):
            subname = f"{name}[{idx}]"
            act_val = actual[idx]
            exp_val = expected[idx]

            if isinstance(act_val, np.ndarray) and isinstance(exp_val, np.ndarray):
                assert_mat_arrays_equal(act_val, exp_val, name=subname)

            elif np.isscalar(act_val) and np.isscalar(exp_val):
                assert (
                    act_val == exp_val
                ), f"{subname}: scalar mismatch {act_val} != {exp_val}"

            else:
                try:
                    np.testing.assert_array_equal(
                        np.asarray(act_val),
                        np.asarray(exp_val),
                        err_msg=f"{subname}: mismatch",
                    )
                except Exception as e:
                    raise AssertionError(f"{subname}: object mismatch: {str(e)}")

    # Case 3: Regular arrays (numeric, string, etc.)
    else:
        try:
            np.testing.assert_array_equal(
                actual, expected, err_msg=f"{name}: array mismatch"
            )
        except AssertionError as e:
            raise AssertionError(f"{name}: {str(e)}")


@pytest.mark.parametrize(
    "var_name, expected",
    variables,
    ids=[v[0] for v in variables],
)
def test_v7_load_struct_cell(var_name, expected):
    loaded_dict = load_from_mat(sc_v73)
    assert_mat_arrays_equal(loaded_dict[var_name], expected)



import os

import numpy as np
import pytest

from matio import load_from_mat

params = [
    (
        {},
        "map1",
    ),
    (
        {
            1: np.array(["a"]),
            2: np.array(["b"]),
        },
        "map2",
    ),
    (
        {
            "a": np.array([[1]]),
            "b": np.array([[2]]),
        },
        "map3",
    ),
    (
        {
            "a": np.array([[1]]),
            "b": np.array([[2]]),
        },
        "map4",
    ),
]


@pytest.mark.parametrize(
    "expected_array, var_name",
    params,
    ids=[
        "map-empty-v7",
        "map-numeric-key-v7",
        "map-char-key-v7",
        "map-string-key-v7",
    ],
)
def test_containermap_load_v7(expected_array, var_name):
    file_path_v7 = os.path.join(os.path.dirname(__file__), "test_map_v7.mat")
    matdict = load_from_mat(file_path_v7, raw_data=False)

    assert var_name in matdict
    for key, val in matdict[var_name].items():
        assert key in expected_array
        assert np.array_equal(val, expected_array[key])


@pytest.mark.parametrize(
    "expected_array, var_name",
    params,
    ids=[
        "map-empty-v7.3",
        "map-numeric-key-v7.3",
        "map-char-key-v7.3",
        "map-string-key-v7.3",
    ],
)
def test_containermap_load_v73(expected_array, var_name):
    file_path_v73 = os.path.join(os.path.dirname(__file__), "test_map_v73.mat")
    matdict = load_from_mat(file_path_v73, raw_data=False)

    assert var_name in matdict
    for key, val in matdict[var_name].items():
        assert key in expected_array
        assert np.array_equal(val, expected_array[key])



"""Reads MCOS subsystem data from MAT files"""

import warnings
from contextlib import contextmanager

import numpy as np
from scipy.io.matlab._mio5 import EmptyStructMarker
from scipy.io.matlab._mio5_params import (
    OPAQUE_DTYPE,
    MatlabFunction,
    MatlabObject,
    MatlabOpaque,
)

from matio.mat_opaque_tools import (
    MatioOpaque,
    convert_mat_to_py,
    convert_py_to_mat,
    guess_class_name,
    mat_to_enum,
)

FILEWRAPPER_INSTANCE = None
OBJECT_CACHE = {}

FILEWRAPPER_VERSION = 4
SYSTEM_BYTE_ORDER = "<u4" if np.little_endian else ">u4"


@contextmanager
def get_matio_context():
    """Context manager for both FileWrapper and object cache"""
    global FILEWRAPPER_INSTANCE, OBJECT_CACHE

    FILEWRAPPER_INSTANCE = None
    OBJECT_CACHE = {}

    try:
        yield
    finally:
        FILEWRAPPER_INSTANCE = None
        OBJECT_CACHE.clear()


def set_file_wrapper():
    """Set global FileWrapper instance"""
    global FILEWRAPPER_INSTANCE
    if FILEWRAPPER_INSTANCE is not None:
        raise RuntimeError(
            "Subsystem data was not cleaned up. Use get_matio_context() to reset"
        )
    FILEWRAPPER_INSTANCE = FileWrapper()
    return FILEWRAPPER_INSTANCE


def get_file_wrapper():
    """Get global FileWrapper instance"""
    if FILEWRAPPER_INSTANCE is None:
        raise RuntimeError(
            "No FileWrapper instance is active. Use get_matio_context() first."
        )
    return FILEWRAPPER_INSTANCE


def get_object_cache():
    """Get global object cache"""
    return OBJECT_CACHE


def is_valid_opaque_object(metadata):
    """Checks if property value is a valid opaque object metadata array"""

    if is_valid_mcos_object(metadata):
        return True

    # TODO: Add checks for other opaque objects

    return False


def is_valid_mcos_object(metadata):
    """Checks if property value is a valid MCOS metadata array"""

    if metadata.dtype.names:
        if "EnumerationInstanceTag" in metadata.dtype.names:
            if (
                metadata[0, 0]["EnumerationInstanceTag"].dtype == np.uint32
                and metadata[0, 0]["EnumerationInstanceTag"].size == 1
                and metadata[0, 0]["EnumerationInstanceTag"] == 0xDD000000
            ):
                return True
        return False

    if not (
        metadata.dtype == np.uint32
        and metadata.ndim == 2
        and metadata.shape == (metadata.shape[0], 1)
        and metadata.size >= 3
    ):
        return False

    if metadata[0, 0] != 0xDD000000:
        return False

    return True


def check_prop_for_opaque(prop):
    """Recursively check if a property value in FileWrapper__ contains opaque objects during load"""

    if not isinstance(prop, np.ndarray):
        return prop

    if is_valid_opaque_object(prop):
        prop = load_opaque_object(prop, classname="", type_system="MCOS")

    elif prop.dtype == object:
        # Iterate through cell arrays
        for idx in np.ndindex(prop.shape):
            cell_item = prop[idx]
            prop[idx] = check_prop_for_opaque(cell_item)

    elif prop.dtype.names:
        # Iterate though struct array
        for idx in np.ndindex(prop.shape):
            for name in prop.dtype.names:
                field_val = prop[idx][name]
                prop[idx][name] = check_prop_for_opaque(field_val)

    return prop


class FileWrapper:
    """Representation class for MATLAB FileWrapper__ data"""

    def __init__(self):
        self.byte_order = SYSTEM_BYTE_ORDER
        self.raw_data = False
        self.add_table_attrs = False
        self._use_strings = True  # Use in future?

        self.version = FILEWRAPPER_VERSION
        self.num_names = 0
        self.region_offsets = None
        self.mcos_names = []
        self.class_id_metadata = []
        self.object_id_metadata = []
        self.saveobj_prop_metadata = []
        self.obj_prop_metadata = []
        self.dynprop_metadata = []
        self._u6_metadata = None  # Unknown Object Metadata
        self._u7_metadata = None  # Unknown Object Metadata
        self.prop_vals_saved = []
        self._c3 = None  # Unknown Class Template
        self._c2 = None  # Unknown Class Template
        self.prop_vals_defaults = None

        # Counters for object creation
        self.saveobj_object_counter = 0
        self.normal_object_counter = 0
        self.class_id_counter = 0
        self.object_id_counter = 0

    def init_load(
        self,
        fwrap_data,
        byte_order=SYSTEM_BYTE_ORDER,
        raw_data=False,
        add_table_attrs=False,
    ):
        """Initializes the FileWrapper instance with metadata from MATLAB FileWrapper__"""
        self.byte_order = "<u4" if byte_order[0] == "<" else ">u4"
        self.raw_data = raw_data
        self.add_table_attrs = add_table_attrs

        fwrap_metadata = fwrap_data[0, 0]

        fromfile_version = np.frombuffer(
            fwrap_metadata, dtype=self.byte_order, count=1, offset=0
        )[0]
        if fromfile_version > self.version:
            raise RuntimeError(
                f"FileWrapper version {fromfile_version} is not supported"
            )

        self.num_names = np.frombuffer(
            fwrap_metadata, dtype=self.byte_order, count=1, offset=4
        )[0]

        self.region_offsets = np.frombuffer(
            fwrap_metadata, dtype=self.byte_order, count=8, offset=8
        )

        # Property and Class Names
        data = fwrap_metadata[40 : self.region_offsets[0]].tobytes()
        raw_strings = data.split(b"\x00")
        self.mcos_names = [s.decode("ascii") for s in raw_strings if s]

        # Class ID Metadata
        self.class_id_metadata = np.frombuffer(
            fwrap_metadata,
            dtype=self.byte_order,
            count=(self.region_offsets[1] - self.region_offsets[0]) // 4,
            offset=self.region_offsets[0],
        )

        # Saveobj Prop Metadata
        self.saveobj_prop_metadata = np.frombuffer(
            fwrap_metadata,
            dtype=self.byte_order,
            count=(self.region_offsets[2] - self.region_offsets[1]) // 4,
            offset=self.region_offsets[1],
        )

        # Object ID Metadata
        self.object_id_metadata = np.frombuffer(
            fwrap_metadata,
            dtype=self.byte_order,
            count=(self.region_offsets[3] - self.region_offsets[2]) // 4,
            offset=self.region_offsets[2],
        )

        # Object Prop Metadata
        self.obj_prop_metadata = np.frombuffer(
            fwrap_metadata,
            dtype=self.byte_order,
            count=(self.region_offsets[4] - self.region_offsets[3]) // 4,
            offset=self.region_offsets[3],
        )

        # Dynamic Prop Metadata
        self.dynprop_metadata = np.frombuffer(
            fwrap_metadata,
            dtype=self.byte_order,
            count=(self.region_offsets[5] - self.region_offsets[4]) // 4,
            offset=self.region_offsets[4],
        )

        # Unknown Region 6 Metadata
        self._u6_metadata = fwrap_metadata[
            self.region_offsets[5] : self.region_offsets[6]
        ]

        # Unknown Region 7 Metadata
        self._u7_metadata = fwrap_metadata[
            self.region_offsets[6] : self.region_offsets[7]
        ]

        self.prop_vals_saved = fwrap_data[2:-3, 0]
        self._c3 = fwrap_data[-3, 0]  # Unknown
        self._c2 = fwrap_data[-2, 0]  # Unknown
        self.prop_vals_defaults = fwrap_data[-1, 0]

    def init_save(self, use_strings=True):
        """Initializes save with metadata for object ID = 0"""

        self._use_strings = use_strings
        self.class_id_metadata.extend([0, 0, 0, 0])
        self.object_id_metadata.extend([0, 0, 0, 0, 0, 0])
        self.saveobj_prop_metadata.extend([0, 0])
        self.obj_prop_metadata.extend([0, 0])
        self.dynprop_metadata.extend([0, 0])

    def get_classname(self, class_id):
        """Extracts class name with namespace qualifier for a given object from its class ID"""

        namespace_idx = self.class_id_metadata[class_id * 4]
        classname_idx = self.class_id_metadata[class_id * 4 + 1]

        if namespace_idx == 0:
            namespace = ""
        else:
            namespace = self.mcos_names[namespace_idx - 1] + "."

        classname = namespace + self.mcos_names[classname_idx - 1]
        return classname

    def get_object_metadata(self, object_id):
        """Extracts object dependency IDs for a given object"""

        class_id, _, _, saveobj_id, normobj_id, dep_id = self.object_id_metadata[
            object_id * 6 : object_id * 6 + 6
        ]
        return class_id, saveobj_id, normobj_id, dep_id

    def get_default_properties(self, class_id):
        """Returns the default properties (as dict) for a given class ID"""

        prop_arr = self.prop_vals_defaults[class_id, 0]
        prop_map = {}
        if prop_arr.dtype.names:
            for prop_name in prop_arr.dtype.names:
                prop_map[prop_name] = check_prop_for_opaque(prop_arr[prop_name][0, 0])

        return prop_map

    def get_property_idxs(self, obj_type_id, saveobj_ret_type):
        """Returns the property field indices for an object"""

        prop_field_idxs = (
            self.saveobj_prop_metadata if saveobj_ret_type else self.obj_prop_metadata
        )

        nfields = 3
        offset = prop_field_idxs[0]
        for _ in range(obj_type_id):
            nprops = prop_field_idxs[offset]
            offset += 1 + nfields * nprops
            offset += offset % 2  # Padding

        nprops = prop_field_idxs[offset]
        offset += 1
        return prop_field_idxs[offset : offset + nprops * nfields].reshape(
            (nprops, nfields)
        )

    def get_saved_properties(self, obj_type_id, saveobj_ret_type):
        """Returns the saved properties (as dict) for an object"""

        save_prop_map = {}

        prop_field_idxs = self.get_property_idxs(obj_type_id, saveobj_ret_type)
        for prop_idx, prop_type, prop_value in prop_field_idxs:
            prop_name = self.mcos_names[prop_idx - 1]
            if prop_type == 0:
                save_prop_map[prop_name] = self.mcos_names[prop_value - 1]
            elif prop_type == 1:
                save_prop_map[prop_name] = check_prop_for_opaque(
                    self.prop_vals_saved[prop_value]
                )
            elif prop_type == 2:
                save_prop_map[prop_name] = prop_value
            else:
                raise ValueError(
                    f"Unknown property type ID:{prop_type} encountered during deserialization"
                )

        return save_prop_map

    def get_dyn_object_id(self, normobj_id):
        """Gets the object ID from normobj ID for dynamicprops objects"""

        num_objects = len(self.object_id_metadata) // 6

        for object_id in range(num_objects):
            block_start = object_id * 6
            block_normobj_id = self.object_id_metadata[block_start + 4]

            if block_normobj_id == normobj_id:
                return object_id

        raise ValueError(f"No dynamicproperty found with ID {normobj_id}")

    def get_dynamic_properties(self, dep_id):
        """Returns dynamicproperties (as dict) for a given object based on dependency ID"""

        offset = self.dynprop_metadata[0]
        for i in range(dep_id):
            nprops = self.dynprop_metadata[offset]
            offset += 1 + nprops
            offset += offset % 2

        ndynprops = self.dynprop_metadata[offset]
        offset += 1
        dyn_prop_type2_ids = self.dynprop_metadata[offset : offset + ndynprops]

        if ndynprops == 0:
            return {}

        dyn_prop_map = {}
        for i, dyn_prop_id in enumerate(dyn_prop_type2_ids):
            dyn_obj_id = self.get_dyn_object_id(dyn_prop_id)
            classname = self.get_classname(dyn_obj_id)
            obj = MatioOpaque(classname=classname, type_system="MCOS")
            obj.properties = self.get_properties(dyn_obj_id)
            dyn_prop_map[f"__dynamic_property__{i + 1}"] = obj

        return dyn_prop_map

    def get_properties(self, object_id):
        """Returns the properties as a dict for a given object ID"""
        if object_id == 0:
            return {}

        class_id, saveobj_id, normobj_id, dep_id = self.get_object_metadata(object_id)
        if saveobj_id != 0:
            saveobj_ret_type = True
            obj_type_id = saveobj_id
        else:
            saveobj_ret_type = False
            obj_type_id = normobj_id

        prop_map = self.get_default_properties(class_id)
        prop_map.update(self.get_saved_properties(obj_type_id, saveobj_ret_type))
        prop_map.update(self.get_dynamic_properties(dep_id))

        return prop_map

    def get_class_id(self, classname):
        """Gets the class ID for a given class name (including namespace)"""

        for class_id in range(1, self.class_id_counter + 1):
            existing_name = self.get_classname(class_id)
            if existing_name == classname:
                return class_id

        self.class_id_counter += 1

        # Add new class ID metadata
        namespace, _, cname = classname.rpartition(".")
        if namespace:
            namespace_idx = self.set_mcos_name(namespace)
        else:
            namespace_idx = 0
        cname_idx = self.set_mcos_name(cname)

        metadata = [namespace_idx, cname_idx, 0, 0]
        self.class_id_metadata.extend(metadata)

        return self.class_id_counter

    def set_mcos_name(self, name):
        """Gets or creates index for a name in mcos_names"""
        try:
            return self.mcos_names.index(name) + 1  # 1-based
        except ValueError:
            # Name doesn't exist, add it
            self.mcos_names.append(name)
            self.num_names += 1
            return self.num_names

    def set_mcos_object_metadata(self, obj, class_id, saveobj_ret_type=False):
        """Sets the metadata for a MATLAB MCOS object"""

        object_cache = get_object_cache()

        obj_key = id(obj)
        cache_id = object_cache.get(obj_key, (0, 0))[0]
        if cache_id == 0:
            cache_id = len(object_cache) + 1
            object_cache[obj_key] = (cache_id, obj)
            new_entry = True
        else:
            new_entry = False

        if not new_entry:
            return cache_id

        prop_map = obj.properties
        self.object_id_counter += 1

        saveobj_object_id = 0
        normal_object_id = 0

        if saveobj_ret_type:
            self.saveobj_object_counter += 1
            saveobj_object_id = self.saveobj_object_counter
        else:
            self.normal_object_counter += 1
            normal_object_id = self.normal_object_counter

        object_id_entry = [class_id, 0, 0, saveobj_object_id, normal_object_id, 0]

        object_id_metadata_pos = len(self.object_id_metadata)
        saveobj_metadata_pos = len(self.saveobj_prop_metadata)
        normobj_metadata_pos = len(self.obj_prop_metadata)

        # Process Properties
        prop_names = list(prop_map.keys())
        num_props = len(prop_names)
        prop_fields = [num_props]

        for prop_name in prop_names:
            field_name_idx = self.set_mcos_name(prop_name)

            prop_value = prop_map[prop_name]
            prop_value = find_matio_opaque(prop_value, in_subsystem=True)

            cell_number_idx = len(self.prop_vals_saved)
            self.prop_vals_saved.append(prop_value)
            prop_fields.extend([field_name_idx, 1, cell_number_idx])
            # TODO:Add support for different property types

        if (num_props * 3 + 1) % 2 != 0:
            prop_fields.append(0)  # Padding

        if saveobj_ret_type:
            self.saveobj_prop_metadata[saveobj_metadata_pos:saveobj_metadata_pos] = (
                prop_fields
            )
        else:
            self.obj_prop_metadata[normobj_metadata_pos:normobj_metadata_pos] = (
                prop_fields
            )

        self.object_id_metadata[object_id_metadata_pos:object_id_metadata_pos] = (
            object_id_entry
        )

        # TODO: Add dynamicprop support
        dynprop_entry = [0, 0]
        self.dynprop_metadata = np.append(
            self.dynprop_metadata, np.array(dynprop_entry, dtype=np.uint32)
        )

        return cache_id

    def write_fwrap_metadata(self):
        """Create FileWrapper Metadata Array"""

        regions = []
        regions.append(np.array([self.version], dtype=np.uint32).view(np.uint8))
        regions.append(np.array([self.num_names], dtype=np.uint32).view(np.uint8))

        # Region offsets (uint32 array)
        region_offsets = np.zeros(8, dtype=np.uint32)
        regions.append(region_offsets.view(np.uint8))

        # Names string (null-terminated)
        names_bytes = (
            b"\x00".join(name.encode("ascii") for name in self.mcos_names) + b"\x00"
        )
        pad_len = (8 - len(names_bytes) % 8) % 8
        names_bytes += b"\x00" * pad_len
        names_uint8 = np.frombuffer(names_bytes, dtype=np.uint8)
        regions.append(names_uint8)
        region_offsets[0] = 40 + names_uint8.size

        # Region 1
        region1 = np.array(self.class_id_metadata, dtype=np.uint32).view(np.uint8)
        regions.append(region1)
        region_offsets[1] = region_offsets[0] + region1.size

        # Region 2
        region2 = (
            np.array(self.saveobj_prop_metadata, dtype=np.uint32).view(np.uint8)
            if len(self.saveobj_prop_metadata) > 2
            else np.empty(0, dtype=np.uint8)
        )
        regions.append(region2)
        region_offsets[2] = region_offsets[1] + region2.size

        # Region 3
        region3 = np.array(self.object_id_metadata, dtype=np.uint32).view(np.uint8)
        regions.append(region3)
        region_offsets[3] = region_offsets[2] + region3.size

        # Region 4
        region4 = (
            np.array(self.obj_prop_metadata, dtype=np.uint32).view(np.uint8)
            if len(self.obj_prop_metadata) > 0
            else np.empty(0, dtype=np.uint8)
        )
        regions.append(region4)
        region_offsets[4] = region_offsets[3] + region4.size

        # Region 5
        region5 = np.array(self.dynprop_metadata, dtype=np.uint32).view(np.uint8)
        regions.append(region5)
        region_offsets[5] = region_offsets[4] + region5.size

        # Region 6
        region6 = np.empty(0, dtype=np.uint8)
        regions.append(region6)
        region_offsets[6] = region_offsets[5] + region6.size

        # Region 7
        region7 = np.zeros(1, np.uint64).view(np.uint8)
        regions.append(region7)
        region_offsets[7] = region_offsets[6] + region7.size

        fwrap_metadata = np.concatenate(regions)
        return fwrap_metadata.reshape(-1, 1)

    def write_subsys_data(self):
        """Create FileWrapper Cell Array"""

        if self.class_id_counter == 0:
            return None

        array_len = 5 + len(self.prop_vals_saved)
        fwrap_data = np.empty((array_len, 1), dtype=object)

        fwrap_data[0, 0] = self.write_fwrap_metadata()
        fwrap_data[1, 0] = np.empty((0, 0), dtype=object)  # Empty Placeholder

        for i, prop_val in enumerate(self.prop_vals_saved):
            fwrap_data[2 + i, 0] = prop_val

        # Write Unknowns
        u3_arr = np.empty((self.class_id_counter + 1, 1), dtype=object)
        for i in range(self.class_id_counter + 1):
            u3_arr[i, 0] = np.empty((1, 0), dtype=object)

        # Write U3 Array
        fwrap_data[-3, 0] = u3_arr
        fwrap_data[-1, 0] = u3_arr

        fwrap_data[-2, 0] = np.zeros(
            shape=(self.class_id_counter + 1, 1), dtype=np.int32
        )
        return fwrap_data


def load_mcos_enumeration(metadata, type_system):
    """Loads MATLAB MCOS enumeration instance array"""

    file_wrapper = get_file_wrapper()

    classname = file_wrapper.get_classname(metadata[0, 0]["ClassName"].item())
    builtin_class_idx = metadata[0, 0]["BuiltinClassName"].item()
    if builtin_class_idx != 0:
        builtin_class_name = file_wrapper.get_classname(builtin_class_idx)
    else:
        builtin_class_name = np.str_("")

    value_names = [
        file_wrapper.mcos_names[val - 1] for val in metadata[0, 0]["ValueNames"].ravel()
    ]

    enum_vals = []
    value_idx = metadata[0, 0]["ValueIndices"]
    mmdata = metadata[0, 0]["Values"]  # Array is N x 1 shape
    if mmdata.size != 0:
        mmdata_map = mmdata[value_idx]
        for val in np.nditer(mmdata_map, flags=["refs_ok"], op_flags=["readonly"]):
            obj_array = load_mcos_object(val.item(), "MCOS")
            enum_vals.append(obj_array)

    if not file_wrapper.raw_data:
        return mat_to_enum(
            enum_vals,
            value_names,
            classname,
            value_idx.shape,
        )

    metadata[0, 0]["BuiltinClassName"] = builtin_class_name
    metadata[0, 0]["ClassName"] = classname
    metadata[0, 0]["ValueNames"] = np.array(value_names).reshape(
        value_idx.shape, order="F"
    )
    metadata[0, 0]["ValueIndices"] = value_idx
    metadata[0, 0]["Values"] = np.array(enum_vals).reshape(value_idx.shape, order="F")

    return MatioOpaque(
        properties=metadata, classname=classname, type_system=type_system
    )


def load_mcos_object(metadata, type_system):
    """Loads MCOS object"""

    file_wrapper = get_file_wrapper()
    object_cache = get_object_cache()

    if metadata.dtype.names:
        return load_mcos_enumeration(metadata, type_system)

    ndims = metadata[1, 0]
    dims = metadata[2 : 2 + ndims, 0]
    nobjects = np.prod(dims)
    object_ids = metadata[2 + ndims : 2 + ndims + nobjects, 0]

    class_id = metadata[-1, 0]
    classname = file_wrapper.get_classname(class_id)

    obj_arr = np.empty((nobjects, 1), dtype=object)

    for i, object_id in enumerate(object_ids):
        if object_id in object_cache:
            obj_arr[i] = object_cache[object_id]
        elif object_id == 0:
            # Empty object, return empty MatioOpaque
            obj = MatioOpaque(classname=classname, type_system=type_system)
            obj.properties = {}
            obj_arr[i] = obj
        else:
            obj = MatioOpaque(classname=classname, type_system=type_system)
            object_cache[object_id] = obj
            obj.properties = file_wrapper.get_properties(object_id)

            if not file_wrapper.raw_data:
                obj = convert_mat_to_py(
                    obj,
                    byte_order=file_wrapper.byte_order,
                    add_table_attrs=file_wrapper.add_table_attrs,
                )

            obj_arr[i, 0] = obj

    if nobjects == 1:
        return obj_arr[0, 0]

    for obj in obj_arr.flat:
        obj.is_array = True
    return obj_arr.reshape(dims, order="F")


def load_opaque_object(metadata, classname, type_system):
    """Loads opaque object"""

    if type_system != "MCOS":
        warnings.warn(
            f"subsystem:load_mcos_object: Loading opaque objects of type "
            f"{type_system} is not supported. Returning metadata"
        )
        obj = MatioOpaque(metadata, classname, type_system)
        return obj

    return load_mcos_object(metadata, type_system)


def create_mcos_metadata(dims, arr_ids, class_id):
    """Creates MCOS metadata array"""

    ndims = len(dims)
    values = [0xDD000000, ndims] + list(dims) + list(arr_ids) + [class_id]
    return np.array(values, dtype=np.uint32).reshape(-1, 1)


def set_object_metadata(obj, saveobj_ret_type=False):
    """Sets metadata for a MatioOpaque object"""

    if isinstance(obj, np.ndarray):
        obj0 = obj.flat[0]
        type_system = obj0.type_system
        classname = obj0.classname
    else:
        type_system = obj.type_system
        classname = obj.classname

    if type_system != "MCOS":
        warnings.warn(
            "subsystem:set_object_metadata: Only MCOS objects are supported currently. This item will be skipped"
        )
        return np.empty((), dtype=np.uint32)

    file_wrapper = get_file_wrapper()

    arr_ids = []
    class_id = file_wrapper.get_class_id(classname)

    if isinstance(obj, np.ndarray):
        # Object Arrays
        for idx in np.ndindex(obj.shape):
            object_id = file_wrapper.set_mcos_object_metadata(obj[idx], class_id)
            arr_ids.append(object_id)
        dims = obj.shape
    else:
        if not isinstance(obj.properties, dict) or obj.classname == "containers.Map":
            obj.properties = convert_py_to_mat(
                obj.properties,
                obj.classname,
            )

        if classname in ("string", "timetable"):
            saveobj_ret_type = True

        object_id = file_wrapper.set_mcos_object_metadata(
            obj, class_id, saveobj_ret_type
        )
        arr_ids.append(object_id)
        dims = (1, 1)

    return create_mcos_metadata(dims, arr_ids, class_id)


def wrap_matlab_opaque(obj):
    """Wrap data in scipy's Matlab Opaque for writing to MAT-files"""

    if isinstance(obj, np.ndarray):
        obj0 = obj.flat[0]
        classname = obj0.classname
        type_system = obj0.type_system
    else:
        classname = obj.classname
        type_system = obj.type_system

    data_new = np.empty((1,), dtype=OPAQUE_DTYPE)
    data_new0 = data_new[0]
    data_new0["_Class"] = classname
    data_new0["_TypeSystem"] = type_system
    data_new0["_ObjectMetadata"] = set_object_metadata(obj)

    data_new = MatlabOpaque(data_new)

    return data_new


def find_matio_opaque(data, in_subsystem=False):
    """Recursively find and wrap MatioOpaque objects in data."""

    if isinstance(data, MatioOpaque):
        if data.classname is None:
            data.classname = guess_class_name(data.properties)
            data.type_system = "MCOS"
        return set_object_metadata(data) if in_subsystem else wrap_matlab_opaque(data)

    if isinstance(data, np.generic):
        # For compatibility
        data = np.asarray(data).reshape(1, 1)

    if isinstance(data, np.ndarray):

        # 2a: datetime
        if np.issubdtype(data.dtype, np.datetime64):
            tmp_obj = MatioOpaque(
                properties=data, classname="datetime", type_system="MCOS"
            )
            return (
                set_object_metadata(tmp_obj)
                if in_subsystem
                else wrap_matlab_opaque(tmp_obj)
            )

        # 2b: duration
        if np.issubdtype(data.dtype, np.timedelta64):
            tmp_obj = MatioOpaque(
                properties=data, classname="duration", type_system="MCOS"
            )
            return (
                set_object_metadata(tmp_obj)
                if in_subsystem
                else wrap_matlab_opaque(tmp_obj)
            )

        if data.dtype.kind in {"U", "S"}:
            if data.dtype.str.endswith("1") or data.size == 1:
                # FIXME: Need a better way to distinguish between char arrays and strings
                return data

            # Read as string array
            tmp_obj = MatioOpaque(
                properties=data, classname="string", type_system="MCOS"
            )
            return (
                set_object_metadata(tmp_obj)
                if in_subsystem
                else wrap_matlab_opaque(tmp_obj)
            )

        # 2b: Cell Arrays
        if data.dtype == object:
            if data.size == 0:
                return data
            if all(
                isinstance(item, MatioOpaque) and item.is_array for item in data.flat
            ):
                return (
                    set_object_metadata(data)
                    if in_subsystem
                    else wrap_matlab_opaque(data)
                )
            for idx in np.ndindex(data.shape):
                data[idx] = find_matio_opaque(data[idx], in_subsystem)
            return data

        # 2c: struct arrays
        if data.dtype.names:
            if set(data.dtype.names) == {"months", "days", "millis"}:
                # calendarDuration
                if not all(
                    np.issubdtype(data[name].item().dtype, np.timedelta64)
                    for name in ("months", "days", "millis")
                ):
                    return data  # If calendarDuration is converted into prop map
                tmp_obj = MatioOpaque(
                    properties=data, classname="calendarDuration", type_system="MCOS"
                )
                return (
                    set_object_metadata(tmp_obj)
                    if in_subsystem
                    else wrap_matlab_opaque(tmp_obj)
                )
            for idx in np.ndindex(data.shape):
                for name in data.dtype.names:
                    data[idx][name] = find_matio_opaque(data[idx][name], in_subsystem)
            return data

        return data

    if isinstance(
        data,
        (
            MatlabObject,
            MatlabFunction,
            MatlabOpaque,
            str,
            list,
            bool,
            int,
            float,
            EmptyStructMarker,
        ),
    ):
        # For scipy compatibility
        # Scipy also accepts tuple, but here it is used for MATLAB dictionary
        return data

    # Everything else: attempt to wrap into MatioOpaque
    classname = guess_class_name(data)
    tmp_obj = MatioOpaque(properties=data, classname=classname, type_system="MCOS")
    return set_object_metadata(tmp_obj) if in_subsystem else wrap_matlab_opaque(tmp_obj)


def create_subsystem_metadata(fwrap_data, java_data=None, handle_data=None):
    """Creates subsystem struct array"""

    if fwrap_data is None and java_data is None and handle_data is None:
        return None

    field_names = []
    field_data = []

    if fwrap_data is not None:
        field_names.append("MCOS")
        data_new = np.empty((1,), dtype=OPAQUE_DTYPE)
        data_new0 = data_new[0]
        data_new0["_Class"] = "FileWrapper__"
        data_new0["_TypeSystem"] = "MCOS"
        data_new0["_ObjectMetadata"] = fwrap_data
        data_new = MatlabOpaque(data_new)
        field_data.append(data_new)

    dtype = [(name, object) for name in field_names]
    subsys_metadata = np.empty((1, 1), dtype=dtype)

    for field_name, data in zip(field_names, field_data):
        subsys_metadata[0, 0][field_name] = data

    return subsys_metadata


def parse_input_dict(mdict):
    """Parses opaque class metadata"""

    file_wrapper = get_file_wrapper()

    for var, data in mdict.items():
        if var[0] == "_":
            continue
        try:
            mdict[var] = find_matio_opaque(data)
        except ValueError as e:
            raise ValueError(f"Error processing variable '{var}': {str(e)}") from e

    fwrap_data = file_wrapper.write_subsys_data()
    subsys_data = create_subsystem_metadata(fwrap_data)
    mdict["__subsystem__"] = subsys_data

    return mdict



import os

import numpy as np
import pytest

from matio import load_from_mat

params = [
    (
        np.array([["2025-04-01T12:00:00"]], dtype="datetime64[ms]").reshape(1, 1),
        "dt1",
    ),
    (
        np.array([["2025-04-01T12:00:00"]], dtype="datetime64[ms]").reshape(1, 1),
        "dt2",
    ),
    (
        np.array(
            [
                [
                    "2025-04-01",
                    "2025-04-03",
                    "2025-04-05",
                    "2025-04-02",
                    "2025-04-04",
                    "2025-04-06",
                ]
            ],
            dtype="datetime64[ms]",
        ).reshape(2, 3),
        "dt3",
    ),
    (
        np.array([], dtype="datetime64[ms]"),
        "dt4",
    ),
]


@pytest.mark.parametrize(
    "expected_array, var_name",
    params,
    ids=[
        "datetime-basic-v7",
        "datetime-timezone-v7",
        "datetime-array-v7",
        "datetime-empty-v7",
    ],
)
def test_datetime_load_v7(expected_array, var_name):
    file_path_v7 = os.path.join(os.path.dirname(__file__), "test_datetime_v7.mat")
    matdict = load_from_mat(file_path_v7, raw_data=False)

    assert var_name in matdict
    np.testing.assert_array_equal(matdict[var_name], expected_array)


@pytest.mark.parametrize(
    "expected_array, var_name",
    params,
    ids=[
        "datetime-basic-v7.3",
        "datetime-timezone-v7.3",
        "datetime-array-v7.3",
        "datetime-empty-v7.3",
    ],
)
def test_datetime_load_v73(expected_array, var_name):
    file_path_v73 = os.path.join(os.path.dirname(__file__), "test_datetime_v73.mat")
    matdict = load_from_mat(file_path_v73, raw_data=False)

    assert var_name in matdict
    np.testing.assert_array_equal(matdict[var_name], expected_array)



"""MATLAB MAT-file version 7.3 (HDF5) reader."""

import warnings

import h5py
import numpy as np
from scipy.io.matlab._mio5_params import MatlabObject
from scipy.io.matlab._mio_utils import (  # pylint: disable=no-name-in-module
    chars_to_strings,
)
from scipy.sparse import coo_matrix, csc_array, issparse

from matio.subsystem import get_matio_context, load_opaque_object, set_file_wrapper


class MatRead7:
    """Reads MAT-file version 7.3 (HDF5) files."""

    def __init__(
        self, file_stream, raw_data=False, add_table_attrs=False, chars_as_strings=True
    ):
        """Initializes the MAT-file reader.
        Parameters
        ----------
            file_stream : h5py.File object
            raw_data : bool, optional
            add_table_attrs : bool, optional
            chars_as_strings : bool, optional
        """
        self.h5stream = file_stream
        self.raw_data = raw_data
        self.add_table_attrs = add_table_attrs
        self.chars_as_strings = chars_as_strings

    def initialize_v73_subsystem(self, byte_order, raw_data, add_table_attrs):
        """Initializes the subsystem for v7.3 MAT-files."""

        subsystem_arr = self.read_struct(self.h5stream["#subsystem#"])
        if "MCOS" in subsystem_arr.dtype.names:
            fwrap_data = subsystem_arr[0, 0]["MCOS"]
            file_wrapper = set_file_wrapper()
            file_wrapper.init_load(fwrap_data, byte_order, raw_data, add_table_attrs)

    def read_int(self, obj, is_empty=0):
        """Reads MATLAB integer arrays from the v7.3 MAT-file."""

        int_decode = obj.attrs.get("MATLAB_int_decode", None)
        if int_decode is not None:
            warnings.warn(
                f"MatRead7:read_int: MATLAB_int_decode {int_decode} is not supported. "
                "This may lead to unexpected behaviour.",
                UserWarning,
            )

        if is_empty:
            return np.empty(shape=obj[()], dtype=obj[()].dtype)

        arr = obj[()]
        if arr.dtype.names:
            # complex number
            real = arr["real"]
            imag = arr["imag"]
            arr = np.empty(shape=arr.shape, dtype=np.complex128)
            arr.real = real
            arr.imag = imag

        return arr.T

    def read_char(self, obj, is_empty=0):
        """Decodes MATLAB char arrays from the v7.3 MAT-file."""

        decode_type = obj.attrs.get("MATLAB_int_decode", None)
        raw = obj[()].T

        if is_empty:
            return np.empty(shape=raw, dtype=np.str_)

        if decode_type == 2:
            codec = "utf-16"
        else:
            warnings.warn(
                f"MatRead7:read_char:MATLAB_int_decode {decode_type} not supported. "
                "This may lead to unexpected behaviour",
                UserWarning,
            )
            codec = "utf-8"

        decoded_arr = np.array(list(raw.tobytes().decode(codec))).reshape(raw.shape)
        if self.chars_as_strings:
            return chars_to_strings(decoded_arr)
        return decoded_arr

    def is_struct_matrix(self, hdf5_group):
        """Check if the HDF5 struct group is a struct matrix or scalar"""

        # Scalar structs are stored directly as members of a group (can be nested)
        # Struct arrays are stored as datasets of HDF5 references
        for key in hdf5_group:
            obj = hdf5_group[key]
            if isinstance(obj, h5py.Group):
                return False

            if isinstance(obj, h5py.Dataset):
                class_name = obj.attrs.get("MATLAB_class", None)
                if class_name is not None:
                    return False

        return True

    def read_struct(self, obj, is_empty=0):
        """Reads MATLAB struct arrays from the v7.3 MAT-file."""

        if is_empty:
            return np.empty(shape=obj[()], dtype=object)

        fields = list(obj.keys())
        field_order = obj.attrs.get("MATLAB_fields", None)
        if field_order is not None:
            # For maximum compatibility with scipy.io
            fields = ["".join(x.astype(str)) for x in field_order]

        if self.is_struct_matrix(obj):
            is_scalar = False
            shape = next(iter(obj.values())).shape
        else:
            is_scalar = True
            shape = (1, 1)

        dt = [(name, object) for name in fields]
        arr = np.empty(shape=shape, dtype=dt)

        for field in obj:
            obj_field = obj[field]
            for idx in np.ndindex(arr.shape):
                if is_scalar:
                    arr[idx][field] = self.read_h5_data(obj_field)
                else:
                    arr[idx][field] = self.read_h5_data(self.h5stream[obj_field[idx]])
        return arr.T

    def read_cell(self, obj, is_empty=0):
        """Reads MATLAB cell arrays from the v7.3 MAT-file."""

        if is_empty:
            return np.empty(shape=obj[()], dtype=object)

        arr = np.empty(shape=obj.shape, dtype=object)
        for idx in np.ndindex(obj.shape):
            ref_data = self.h5stream[obj[idx]]
            arr[idx] = self.read_h5_data(ref_data)
        return arr.T

    def read_sparse(self, obj, nrows):
        """Reads MATLAB sparse arrays from the v7.3 MAT-file."""

        jc = obj["jc"][()]
        ncols = jc.size - 1

        if "data" in obj:
            # Exists only if sparse matrix contains non-zero elements
            data = self.read_int(obj["data"])
            ir = obj["ir"][()]
        else:
            data = np.array([], dtype=np.float64)
            ir = np.array([], dtype=np.int32)

        return csc_array((data, ir, jc), shape=(nrows, ncols))

    def read_function_handle(self, obj, object_decode):
        """Reads MATLAB function handles from the v7.3 MAT-file."""

        # Uses object_decode = 1 probably

        if object_decode == 1:
            return self.read_struct(obj)

        raise NotImplementedError(
            f"Function handle object_decode {object_decode} not supported. Only 1 is supported."
        )

    def read_object(self, obj, class_name):
        """Reads mxOBJECT_CLASS variables from the v7.3 MAT-file."""

        class_name = class_name.decode("utf-8")
        fields = self.read_struct(obj)

        return MatlabObject(fields, class_name)

    def get_type_system(self, class_name):
        """Gets the type system for the given class name."""

        # Possibly, MATLAB decodes internally based on class name
        # This function is just guess work as of now
        class_name = class_name.decode("utf-8")
        if class_name.startswith("java.") or class_name.startswith("com."):
            type_system = "java"
        elif class_name.startswith("COM."):
            type_system = "handle"
        else:
            type_system = "MCOS"

        return type_system

    def read_opaque(self, obj, object_decode, is_empty=0):
        """Reads MATLAB opaque objects from the v7.3 MAT-file."""

        class_name = obj.attrs.get("MATLAB_class", None)
        if class_name == b"FileWrapper__":
            return self.read_cell(obj)

        if object_decode == 2:
            # object_decode is a flag indicating how to decode objects
            # Object Decode = 1 -> Function Handle (possibly)
            # Object Decode = 2 -> mxOBJECT_CLASS
            # Object Decode = 3 -> mxOPAQUE_CLASS
            return self.read_object(obj, class_name)

        if is_empty:
            return np.empty(shape=obj[()], dtype=object)

        type_system = self.get_type_system(class_name)

        # Check Enumeration Instances
        fields = obj.attrs.get("MATLAB_fields", None)
        if fields is not None:
            fields = ["".join(x.astype(str)) for x in fields]

            if "EnumerationInstanceTag" in fields:
                metadata = self.read_struct(obj)
                if metadata[0, 0]["EnumerationInstanceTag"] != 0xDD000000:
                    return metadata
            else:
                metadata = obj[()].T
        else:
            metadata = obj[()].T
        return load_opaque_object(metadata, class_name, type_system)

    def read_h5_data(self, obj):
        """Reads data from the HDF5 object."""

        matlab_class = obj.attrs.get("MATLAB_class", None)
        is_empty = obj.attrs.get("MATLAB_empty", 0)
        object_decode = obj.attrs.get("MATLAB_object_decode", -1)
        matlab_sparse = obj.attrs.get("MATLAB_sparse", -1)

        if matlab_sparse >= 0:
            arr = self.read_sparse(obj, matlab_sparse)
        elif matlab_class == b"char":
            arr = self.read_char(obj, is_empty)
        elif matlab_class == b"logical":
            arr = obj[()].T.astype(np.bool_)
        elif matlab_class in (
            b"int8",
            b"uint8",
            b"int16",
            b"uint16",
            b"int32",
            b"uint32",
            b"int64",
            b"uint64",
            b"single",
            b"double",
        ):
            arr = self.read_int(obj, is_empty)
        elif matlab_class == b"struct":
            arr = self.read_struct(obj, is_empty)
        elif matlab_class == b"cell":
            arr = self.read_cell(obj, is_empty)
        elif matlab_class == b"function_handle":
            arr = self.read_function_handle(obj, object_decode)
        elif matlab_class == b"canonical empty" and is_empty:
            arr = np.empty(shape=(0, 0), dtype=object)
        elif object_decode >= 0:
            arr = self.read_opaque(obj, object_decode, is_empty)
        else:
            raise NotImplementedError(
                f"MATLAB class {matlab_class} not supported", UserWarning
            )

        return arr

    def get_variables(self, variable_names, byte_order, raw_data, add_table_attrs):
        """Reads variables from the HDF5 file."""
        if isinstance(variable_names, str):
            variable_names = [variable_names]
        elif variable_names is not None:
            variable_names = list(variable_names)

        mdict = {}
        mdict["__globals__"] = []

        with get_matio_context():

            if "#subsystem#" in self.h5stream:
                self.initialize_v73_subsystem(byte_order, raw_data, add_table_attrs)

            for var in self.h5stream:
                obj = self.h5stream[var]
                if var in ("#refs#", "#subsystem#"):
                    continue
                if variable_names is not None and var not in variable_names:
                    continue
                try:
                    data = self.read_h5_data(obj)
                except Exception as err:
                    raise ValueError(f"Error reading variable {var}: {err}") from err
                mdict[var] = data
                is_global = obj.attrs.get("MATLAB_global", 0)
                if is_global:
                    mdict["__globals__"].append(var)
                if variable_names is not None:
                    variable_names.remove(var)
                    if len(variable_names) == 0:
                        break

        return mdict


def read_file_header(file_path):
    """Reads the file header of the MAT-file."""
    with open(file_path, "rb") as f:
        f.seek(0)
        hdr = f.read(128)
        v_major = hdr[125] if hdr[126] == b"I"[0] else hdr[124]
        v_minor = hdr[124] if hdr[126] == b"I"[0] else hdr[125]
        byte_order = "<" if hdr[126] == b"I"[0] else ">"

        hdict = {}
        hdict["__header__"] = hdr[0:116].decode("utf-8").strip(" \t\n\000")
        hdict["__version__"] = f"{v_major}.{v_minor}"
    return hdict, byte_order


def read_matfile7(
    file_path,
    raw_data=False,
    add_table_attrs=False,
    spmatrix=True,
    byte_order=None,  # pylint: disable=unused-argument
    mat_dtype=False,  # pylint: disable=unused-argument
    chars_as_strings=True,
    verify_compressed_data_integrity=True,  # pylint: disable=unused-argument
    variable_names=None,
):
    """Reads MAT-file version 7.3 (HDF5) files.
    Parameters
    ----------
        file_path : str
            Path to the MAT-file.
        raw_data : bool, optional
            If True, returns raw data for MATLAB object instances. Default is False.
        add_table_attrs : bool, optional
            If True, adds table attributes to Pandas DataFrames. Default is False.
        spmatrix : bool, optional
            If True, converts sparse matrices to COO format. Default is True.
        chars_as_strings : bool, optional
            If True, converts character arrays to strings. Default is True.
        variable_names : list of str or str, optional
            Names of variables to read from the MAT-file. Default is None (reads all variables).
    """

    matfile_dict, byte_order = read_file_header(file_path)
    f = h5py.File(file_path, "r")
    mat_reader = MatRead7(f, raw_data, add_table_attrs, chars_as_strings)
    try:
        mdict = mat_reader.get_variables(
            variable_names, byte_order, raw_data, add_table_attrs
        )
    finally:
        f.close()

    if spmatrix:
        for name, var in list(matfile_dict.items()):
            if issparse(var):
                matfile_dict[name] = coo_matrix(var)

    matfile_dict.update(mdict)
    return matfile_dict



import os

import numpy as np
import pytest

from matio import load_from_mat

params = [
    (np.array([10]).reshape(1, 1), "var_int"),
    (
        np.array(
            [
                np.array(["String in Cell"]).reshape(1, 1),
            ]
        ).reshape(1, 1),
        "var_cell",
    ),
    (
        np.array(
            [[(np.array(["String in Struct"]).reshape(1, 1),)]],
            dtype=[("MyField", "O")],
        ),
        "var_struct",
    ),
]


@pytest.mark.parametrize(
    "expected_array, var_name",
    params,
    ids=["simple-string-v7", "string-in-cell-v7", "string-in-struct-v7"],
)
def test_simple_load_v7(expected_array, var_name):
    file_path_v7 = os.path.join(os.path.dirname(__file__), "test_simple_v7.mat")
    matdict = load_from_mat(file_path_v7, raw_data=False)

    assert var_name in matdict
    assert isinstance(matdict[var_name], np.ndarray)
    np.testing.assert_array_equal(matdict[var_name], expected_array)


@pytest.mark.parametrize(
    "expected_array, var_name",
    params,
    ids=["simple-string-v7.3", "string-in-cell-v7.3", "string-in-struct-v7.3"],
)
def test_simple_load_v73(expected_array, var_name):
    file_path_v73 = os.path.join(os.path.dirname(__file__), "test_simple_v73.mat")
    matdict = load_from_mat(file_path_v73, raw_data=False)

    assert var_name in matdict
    assert isinstance(matdict[var_name], np.ndarray)
    np.testing.assert_array_equal(matdict[var_name], expected_array)



import os
from enum import Enum

import numpy as np
import pytest

from matio import load_from_mat


class EnumClass(Enum):
    """Enum class for testing purposes."""

    enum1 = {"val": np.array([1.0]).reshape(1, 1)}
    enum2 = {"val": np.array([2.0]).reshape(1, 1)}
    enum3 = {"val": np.array([3.0]).reshape(1, 1)}
    enum4 = {"val": np.array([4.0]).reshape(1, 1)}
    enum5 = {"val": np.array([5.0]).reshape(1, 1)}
    enum6 = {"val": np.array([6.0]).reshape(1, 1)}


class EnumClass2(Enum):
    """Enum class for testing"""

    enum1 = {"uint32.Data": np.array([1], dtype=np.uint32).reshape(1, 1)}


params_base = [
    (
        np.array(
            [
                [
                    EnumClass.enum1,
                ]
            ]
        ),
        "enum_base",
    ),
    (
        np.array(
            [
                EnumClass.enum1,
                EnumClass.enum2,
                EnumClass.enum3,
                EnumClass.enum4,
                EnumClass.enum5,
                EnumClass.enum6,
            ]
        ).reshape(2, 3, order="F"),
        "enum_arr",
    ),
    (
        np.array(
            [
                [
                    EnumClass2.enum1,
                ]
            ]
        ),
        "enum_uint32",
    ),
]


@pytest.mark.parametrize(
    "expected_array, var_name",
    params_base,
    ids=["enum-base-v7", "enum-array-v7", "enum-derived-v7"],
)
def test_enum_load_v7(expected_array, var_name):
    file_path = os.path.join(os.path.dirname(__file__), "test_enum_v7.mat")
    matdict = load_from_mat(file_path, raw_data=False)

    assert var_name in matdict
    assert matdict[var_name].shape == expected_array.shape
    assert matdict[var_name].dtype == expected_array.dtype

    for idx in np.ndindex(expected_array.shape):
        actual_member = matdict[var_name][idx]
        expected_member = expected_array[idx]
        assert actual_member.name == expected_member.name
        assert actual_member.value == expected_member.value


@pytest.mark.parametrize(
    "expected_array, var_name",
    params_base,
    ids=["enum-base-v7.3", "enum-array-v7.3", "enum-derived-v7.3"],
)
def test_enum_load_v73(expected_array, var_name):
    file_path = os.path.join(os.path.dirname(__file__), "test_enum_v73.mat")
    matdict = load_from_mat(file_path, raw_data=False)

    # Output format
    assert var_name in matdict
    assert matdict[var_name].shape == expected_array.shape
    assert matdict[var_name].dtype == expected_array.dtype

    for idx in np.ndindex(expected_array.shape):
        actual_member = matdict[var_name][idx]
        expected_member = expected_array[idx]
        assert actual_member.name == expected_member.name
        assert actual_member.value == expected_member.value



import os

import numpy as np
import pytest
from scipy.io import loadmat

from matio import load_from_mat

basic_v7 = os.path.join(os.path.dirname(__file__), "basic_v7.mat")
basic_v73 = os.path.join(os.path.dirname(__file__), "basic_v73.mat")

v7_data = loadmat(basic_v7, mat_dtype=True)
variables = [(key, v7_data[key]) for key in v7_data.keys() if not key.startswith("__")]


@pytest.mark.parametrize(
    "var_name, expected",
    variables,
    ids=[v[0] for v in variables],
)
def test_v73_load_basic(var_name, expected):
    loaded_dict = load_from_mat(basic_v73)
    if isinstance(expected, np.ndarray):
        np.testing.assert_array_equal(loaded_dict[var_name], expected)



"""Reads MAT-files v7 to v7.2 (MAT-file 5) and extracts variables including MATLAB objects"""

import zlib
from io import BytesIO

import numpy as np
from scipy.io import loadmat
from scipy.io.matlab._mio5 import MatFile5Reader, MatFile5Writer
from scipy.io.matlab._mio5_params import MatlabOpaque

from matio.subsystem import (
    get_matio_context,
    load_opaque_object,
    parse_input_dict,
    set_file_wrapper,
)


def new_opaque_object(arr):
    """Creates a new MatioOpaque object in place of a MatlabOpaque array"""

    metadata = arr["_ObjectMetadata"].item()
    classname = arr["_Class"].item()
    type_system = arr["_TypeSystem"].item()

    obj = load_opaque_object(metadata, classname, type_system)
    return obj


def find_matlab_opaque(arr):
    """Iterate through scipy.loadmat return value to find and replace MatlabOpaque objects"""

    if not isinstance(arr, np.ndarray):
        return arr

    if isinstance(arr, MatlabOpaque):
        arr = new_opaque_object(arr)

    elif arr.dtype == object:
        # Iterate through cell arrays
        for idx in np.ndindex(arr.shape):
            cell_item = arr[idx]
            arr[idx] = find_matlab_opaque(cell_item)

    elif arr.dtype.names:
        # Iterate though struct array
        for idx in np.ndindex(arr.shape):
            for name in arr.dtype.names:
                field_val = arr[idx][name]
                arr[idx][name] = find_matlab_opaque(field_val)

    return arr


def read_subsystem(
    ssdata,
    byte_order,
    mat_dtype,
    verify_compressed_data_integrity,
):
    """Reads subsystem data as a MAT-file stream"""
    ss_stream = BytesIO(ssdata)

    ss_stream.seek(8)  # Skip subsystem header
    subsystem_reader = MatFile5Reader(
        ss_stream,
        byte_order=byte_order,
        mat_dtype=mat_dtype,
        verify_compressed_data_integrity=verify_compressed_data_integrity,
    )
    subsystem_reader.initialize_read()
    try:
        hdr, _ = subsystem_reader.read_var_header()
        res = subsystem_reader.read_var_array(hdr, process=False)
    except Exception as err:
        raise ValueError(f"Error reading subsystem data: {err}") from err

    return res


def read_matfile5(
    file_path,
    raw_data=False,
    add_table_attrs=False,
    spmatrix=True,
    byte_order=None,
    mat_dtype=False,
    chars_as_strings=True,
    verify_compressed_data_integrity=True,
    variable_names=None,
):
    """Loads variables from MAT-file < v7.3
    Inputs
        1. raw_data (bool): Whether to return raw data for objects
        2. add_table_attrs (bool): Add attributes to pandas DataFrame
        3. spmatrix (bool): Additional arguments for scipy.io.loadmat
        4. byte_order (str): Endianness
        5. mat_dtype (bool): Whether to load MATLAB data types
        6. chars_as_strings (bool): Whether to load character arrays as strings
        8. verify_compressed_data_integrity (bool): Whether to verify compressed data integrity
        9. variable_names (list): List of variable names to load
    Returns:
        1. matfile_dict (dict): Dictionary of loaded variables
    """
    if variable_names is not None:
        if isinstance(variable_names, str):
            variable_names = [variable_names, "__function_workspace__"]
        elif not isinstance(variable_names, list):
            raise TypeError("variable_names must be a string or a list of strings")
        else:
            variable_names.append("__function_workspace__")

    matfile_dict = loadmat(
        file_path,
        spmatrix=spmatrix,
        byte_order=byte_order,
        mat_dtype=mat_dtype,
        chars_as_strings=chars_as_strings,
        verify_compressed_data_integrity=verify_compressed_data_integrity,
        variable_names=variable_names,
    )
    ssdata = matfile_dict.pop("__function_workspace__", None)
    if ssdata is None:
        # No subsystem data in file
        return matfile_dict

    byte_order = "<" if ssdata[0, 2] == b"I"[0] else ">"

    ss_array = read_subsystem(
        ssdata,
        byte_order,
        mat_dtype,
        verify_compressed_data_integrity,
    )

    if "MCOS" in ss_array.dtype.names:
        if ss_array[0, 0]["MCOS"]["_Class"].item() != "FileWrapper__":
            raise ValueError("Missing-FileWrapper__: Cannot load MATLAB MCOS object")

    fwrap_data = ss_array[0, 0]["MCOS"][0]["_ObjectMetadata"]

    with get_matio_context():

        file_wrapper = set_file_wrapper()
        file_wrapper.init_load(fwrap_data, byte_order, raw_data, add_table_attrs)

        for var, data in matfile_dict.items():
            matfile_dict[var] = find_matlab_opaque(data)

    return matfile_dict


def save_matfile5(
    file_stream,
    mdict,
    do_compression=True,
    global_vars=None,
    oned_as="row",
    use_strings=True,
):
    """Saves variables to MAT-file 5 format (< 7.3)"""

    with get_matio_context():
        file_wrapper = set_file_wrapper()
        file_wrapper.init_save(use_strings)

        mdict = parse_input_dict(mdict)
    subsys_data = mdict.pop("__subsystem__", None)

    scipy_writer = MatFile5Writer(
        file_stream,
        do_compression=do_compression,
        global_vars=global_vars,
        oned_as=oned_as,
    )
    scipy_writer.put_variables(mdict)

    if subsys_data is None:
        return 0

    subsys_data = {"__subsystem__": subsys_data}
    subsys_offset = file_stream.tell()

    subsys_stream = BytesIO()
    pos_mimatrix, pos_ndims, pos_nbytes = write_subsystem_headers(subsys_stream)
    pos_subsys_start = subsys_stream.tell()

    if np.little_endian:
        subsys_stream.write(
            np.array([0, 1, ord("I"), ord("M"), 0, 0, 0, 0], dtype=np.uint8).tobytes()
        )
    else:
        subsys_stream.write(
            np.array([1, 0, ord("M"), ord("I"), 0, 0, 0, 0], dtype=np.uint8).tobytes()
        )

    scipy_writer.do_compression = False
    scipy_writer.file_stream = subsys_stream
    scipy_writer.put_variables(subsys_data)

    len_subsys_data = subsys_stream.tell() - pos_subsys_start
    update_subsystem_headers(
        subsys_stream, pos_mimatrix, pos_ndims, pos_nbytes, len_subsys_data
    )

    if do_compression:
        out_str = zlib.compress(subsys_stream.getvalue())
        tag = np.array([15, len(out_str)], dtype=np.uint32).view(np.uint8)
        file_stream.write(tag.tobytes())
        file_stream.write(out_str)
    else:
        file_stream.write(subsys_stream.getvalue())

    return subsys_offset


def write_subsystem_headers(file_stream):
    """Write subsystem headers to the file stream"""

    file_stream.write(np.array([14, 0], dtype=np.uint32).tobytes())  # miMATRIX Header
    pos_mimatrix = file_stream.tell() - 4

    file_stream.write(np.array([6, 8, 9, 0], dtype=np.uint32).tobytes())  # Array Flag
    file_stream.write(
        np.array([5, 8, 1, 0], dtype=np.uint32).tobytes()
    )  # Dimensions Flag
    pos_ndims = file_stream.tell() - 4

    file_stream.write(np.array([1, 0], dtype=np.uint32).tobytes())  # Array Name
    file_stream.write(np.array([2, 0], dtype=np.uint32).tobytes())  # Array Value Header
    pos_nbytes = file_stream.tell() - 4

    return pos_mimatrix, pos_ndims, pos_nbytes


def update_subsystem_headers(
    file_stream, pos_mimatrix, pos_ndims, pos_nbytes, len_subsys_data
):
    """Update subsystem headers in the file stream"""

    mi_matrix_nbytes = len_subsys_data + 48  # Include Data element headers

    file_stream.seek(pos_mimatrix)
    file_stream.write(np.array([mi_matrix_nbytes], dtype=np.uint32).tobytes())

    file_stream.seek(pos_ndims)
    file_stream.write(np.array([len_subsys_data], dtype=np.uint32).tobytes())

    file_stream.seek(pos_nbytes)
    file_stream.write(np.array([len_subsys_data], dtype=np.uint32).tobytes())

    file_stream.seek(0, 2)  # Seek to end



"""Conversion utilities for matio"""

from .matmap import (
    containermap_to_mat,
    dictionary_to_mat,
    mat_to_containermap,
    mat_to_dictionary,
)
from .matstring import mat_to_string, string_to_mat
from .mattables import (
    categorical_to_mat,
    mat_to_categorical,
    mat_to_table,
    mat_to_timetable,
    table_to_mat,
    timetable_to_mat,
)
from .mattimes import (
    calendarduration_to_mat,
    datetime_to_mat,
    duration_to_mat,
    mat_to_calendarduration,
    mat_to_datetime,
    mat_to_duration,
)

__all__ = [
    "mat_to_containermap",
    "mat_to_dictionary",
    "mat_to_string",
    "mat_to_categorical",
    "mat_to_table",
    "mat_to_timetable",
    "mat_to_calendarduration",
    "mat_to_datetime",
    "mat_to_duration",
    "string_to_mat",
    "datetime_to_mat",
    "duration_to_mat",
    "calendarduration_to_mat",
    "dictionary_to_mat",
    "containermap_to_mat",
    "table_to_mat",
    "timetable_to_mat",
    "categorical_to_mat",
]



"""Convert MATLAB objects to Python compatible objects"""

import warnings
from enum import Enum

import numpy as np
from pandas import Categorical, DataFrame

from matio.utils import (
    calendarduration_to_mat,
    categorical_to_mat,
    containermap_to_mat,
    datetime_to_mat,
    dictionary_to_mat,
    duration_to_mat,
    mat_to_calendarduration,
    mat_to_categorical,
    mat_to_containermap,
    mat_to_datetime,
    mat_to_dictionary,
    mat_to_duration,
    mat_to_string,
    mat_to_table,
    mat_to_timetable,
    string_to_mat,
    table_to_mat,
    timetable_to_mat,
)

MAT_TO_PY = {
    "calendarDuration": mat_to_calendarduration,
    "categorical": mat_to_categorical,
    "containers.Map": mat_to_containermap,
    "datetime": mat_to_datetime,
    "dictionary": mat_to_dictionary,
    "duration": mat_to_duration,
    "string": mat_to_string,
    "table": mat_to_table,
    "timetable": mat_to_timetable,
}

PY_TO_MAT = {
    "calendarDuration": calendarduration_to_mat,
    "categorical": categorical_to_mat,
    "containers.Map": containermap_to_mat,
    "datetime": datetime_to_mat,
    "dictionary": dictionary_to_mat,
    "duration": duration_to_mat,
    "string": string_to_mat,
    "table": table_to_mat,
    "timetable": timetable_to_mat,
}


class MatioOpaque:
    """Represents a MATLAB opaque object"""

    def __init__(self, properties=None, classname=None, type_system="MCOS"):

        self.classname = classname
        self.type_system = type_system
        self.properties = properties
        self.is_array = False

    def __repr__(self):
        return f"MatioOpaque(classname={self.classname})"

    def __eq__(self, other):
        if isinstance(other, MatioOpaque):
            return self.properties == other.properties
        return self.properties == other


def convert_mat_to_py(obj, **kwargs):
    """Converts a MATLAB object to a Python object"""
    convert_func = MAT_TO_PY.get(obj.classname)
    if convert_func is not None:
        return convert_func(
            obj.properties,
            byte_order=kwargs.get("byte_order", None),
            add_table_attrs=kwargs.get("add_table_attrs", None),
        )
    return obj


def convert_py_to_mat(properties, classname):
    """Convert a Python object to a MATLAB object"""

    convert_func = PY_TO_MAT.get(classname)
    if convert_func is None:
        warnings.warn(
            f"convert_py_to_mat: Conversion of {type(properties)} into MATLAB type "
            f"{classname} is not yet implemented. This will be skipped"
        )
        return {}

    return convert_func(properties)


def mat_to_enum(values, value_names, class_name, shapes):
    """Converts MATLAB enum to Python enum"""

    enum_class = Enum(
        class_name,
        {name: val.properties for name, val in zip(value_names, values)},
    )

    enum_members = [enum_class(val.properties) for val in values]
    return np.array(enum_members, dtype=object).reshape(shapes, order="F")


def guess_class_name(properties):
    """Guess the class name based on properties"""

    if properties is None:
        return properties

    classname = None
    if isinstance(properties, DataFrame):
        if isinstance(properties.index.values, np.ndarray) and (
            np.issubdtype(properties.index.values.dtype, np.datetime64)
            or np.issubdtype(properties.index.values.dtype, np.timedelta64)
        ):
            classname = "timetable"
        else:
            classname = "table"
    elif isinstance(properties, Categorical):
        classname = "categorical"
    elif isinstance(properties, dict):
        classname = "containers.Map"
    elif isinstance(properties, tuple):
        classname = "dictionary"
    elif isinstance(properties, np.ndarray):
        if set(properties.dtype.names) == {"months", "days", "millis"}:
            classname = "calendarDuration"
        elif np.issubdtype(properties.dtype, np.datetime64):
            classname = "datetime"
        elif np.issubdtype(properties.dtype, np.timedelta64):
            classname = "duration"
        elif properties.dtype.kind in ("U", "S"):
            classname = "string"

    if classname is None:
        raise ValueError(
            f"Unable to determine MATLAB class equivalent for properties of "
            f"type {type(properties)}. Provide it explicitly."
        )

    return classname



"""Utility functions for converting MATLAB containerMap and Dictionary"""

import warnings

import numpy as np

MAT_DICT_VERSION = 1


def mat_to_containermap(props, **_kwargs):
    """Converts MATLAB container.Map to Python dictionary"""
    comps = props.get("serialization", None)
    if comps is None:
        return props

    ks = comps[0, 0]["keys"]
    vals = comps[0, 0]["values"]

    result = {}
    for i in range(ks.shape[1]):
        key = ks[0, i].item()
        val = vals[0, i]
        result[key] = val

    return result


def mat_to_dictionary(props, **_kwargs):
    """Converts MATLAB dictionary to Python list of tuples"""
    # List of tuples as Key-Value pairs can be any datatypes

    comps = props.get("data", None)
    if comps is None:
        return props

    ver = int(comps[0, 0]["Version"].item())
    if ver != MAT_DICT_VERSION:
        warnings.warn(
            f"mat_to_dictionary: Only v{MAT_DICT_VERSION} MATLAB dictionaries are supported. Got v{ver}",
            UserWarning,
        )
        return props

    ks = comps[0, 0]["Key"]
    vals = comps[0, 0]["Value"]

    return (ks, vals)


def dictionary_to_mat(props):
    """Converts a Python dictionary to MATLAB dictionary"""
    if not (isinstance(props, tuple) and len(props) == 2):
        raise TypeError("Expected tuple of (key, value)")

    keys, values = props

    def is_valid_array(obj):
        return (
            isinstance(obj, np.ndarray) or getattr(obj, "classname", None) == "string"
        )

    if not is_valid_array(keys) or not is_valid_array(values):
        raise TypeError(
            "Keys must be a numpy array or MatioOpaque with classname='string'"
        )

    dtype = [
        ("Version", object),
        ("IsKeyCombined", object),
        ("IsValueCombined", object),
        ("Key", object),
        ("Value", object),
    ]
    data_arr = np.empty((1, 1), dtype=dtype)

    data_arr["Key"][0, 0] = keys
    data_arr["Value"][0, 0] = values
    data_arr["Version"][0, 0] = np.uint64(MAT_DICT_VERSION)
    data_arr["IsKeyCombined"][0, 0] = np.bool_(True)
    data_arr["IsValueCombined"][0, 0] = np.bool_(True)

    prop_map = {
        "data": data_arr,
    }

    return prop_map


def containermap_to_mat(props):
    """Converts a Python dictionary to MATLAB container.Map"""
    if not isinstance(props, dict):
        raise TypeError(f"Expected dict, got {type(props)}")
    if "serialization" in props:
        warnings.warn(
            "containermap_to_mat: Key 'Serialization' was found in the map. This clashes with an expected MATLAB "
            "keyword and will be treated as a raw property map."
        )
        return props

    keys = list(props.keys())
    vals = list(props.values())

    val_arr = np.empty((1, len(vals)), dtype=object)
    keys_arr = np.empty((1, len(keys)), dtype=object)
    val_arr[0, :] = vals
    keys_arr[0, :] = keys

    if all(isinstance(k, str) for k in keys):
        key_type = "char"
    elif all(isinstance(k, int) for k in keys):
        key_type = "uint64"
        # Defaulting to highest precision as Python doesn't differentiate
    else:
        key_type = "double"

    value_dtypes = {v.dtype for v in vals if isinstance(v, np.ndarray)}
    if len(value_dtypes) == 1:
        uniformity = True
        dtype_map = {
            np.dtype("float64"): "double",
            np.dtype("float32"): "single",
            np.dtype("bool"): "logical",
            np.dtype("int8"): "int8",
            np.dtype("uint8"): "uint8",
            np.dtype("int16"): "int16",
            np.dtype("uint16"): "uint16",
            np.dtype("int32"): "int32",
            np.dtype("uint32"): "uint32",
            np.dtype("int64"): "int64",
            np.dtype("uint64"): "uint64",
        }
        value_type = dtype_map.get(next(iter(value_dtypes)), "any")
    else:
        uniformity = np.bool_(False)
        value_type = "any"

    ser_dtype = [
        ("keys", object),
        ("values", object),
        ("uniformity", object),
        ("keyType", object),
        ("valueType", object),
    ]
    serialization = np.empty((1, 1), dtype=ser_dtype)
    serialization[0, 0] = (keys_arr, val_arr, uniformity, key_type, value_type)
    prop_map = {
        "serialization": serialization,
    }
    return prop_map



import os

import numpy as np
import pytest

from matio import load_from_mat
from matio.mat_opaque_tools import MatioOpaque

params_base = [
    (
        {
            "a": np.array([]).reshape(0, 0),
            "b": np.array([]).reshape(0, 0),
            "c": np.array([]).reshape(0, 0),
        },
        "obj1",
    ),
    (
        {
            "a": np.array([10]).reshape(1, 1),
            "b": np.array([20]).reshape(1, 1),
            "c": np.array([30]).reshape(1, 1),
        },
        "obj2",
    ),
    (
        {
            "a": np.array([]).reshape(0, 0),
            "b": np.array([10]).reshape(1, 1),
            "c": np.array([30]).reshape(1, 1),
        },
        "obj3",
    ),
    (
        np.tile(
            np.array(
                [
                    {
                        "a": np.array([10]).reshape(1, 1),
                        "b": np.array([20]).reshape(1, 1),
                        "c": np.array([30]).reshape(1, 1),
                    }
                ],
                dtype=object,
            ),
            (2, 3),
        ),
        "obj6",
    ),
    (
        {
            "a": np.array(["Default String"]).reshape(1, 1),
            "b": np.array([10]).reshape(1, 1),
            "c": np.array([30]).reshape(1, 1),
        },
        "obj7",
    ),
]


@pytest.mark.parametrize(
    "expected_array, var_name",
    params_base,
    ids=[
        "object-without-constructor-v7",
        "object-with-constructor-v7",
        "object-with-default-v7",
        "object-array-v7",
        "object-in-default-property-v7",
    ],
)
def test_user_defined_load_v7(expected_array, var_name):
    file_path = os.path.join(os.path.dirname(__file__), "test_user_defined_v7.mat")
    matdict = load_from_mat(file_path, raw_data=False)

    assert var_name in matdict
    if isinstance(matdict[var_name], MatioOpaque):
        for key, value in matdict[var_name].properties.items():
            np.testing.assert_array_equal(value, expected_array[key])

    if isinstance(matdict[var_name], np.ndarray):
        np.testing.assert_array_equal(matdict[var_name], expected_array)


@pytest.mark.parametrize(
    "expected_array, var_name",
    params_base,
    ids=[
        "object-without-constructor-v7.3",
        "object-with-constructor-v7.3",
        "object-with-default-v7.3",
        "object-array-v7.3",
        "object-in-default-property-v7.3",
    ],
)
def test_user_defined_load_v73(expected_array, var_name):
    file_path = os.path.join(os.path.dirname(__file__), "test_user_defined_v73.mat")
    matdict = load_from_mat(file_path, raw_data=False)

    assert var_name in matdict
    if isinstance(matdict[var_name], MatioOpaque):
        for key, value in matdict[var_name].properties.items():
            np.testing.assert_array_equal(value, expected_array[key])

    if isinstance(matdict[var_name], np.ndarray):
        np.testing.assert_array_equal(matdict[var_name], expected_array)



import os

import numpy as np
import pytest

from matio import load_from_mat

params = [
    (
        np.array(["Hello"], dtype=np.str_).reshape(1, 1),
        "s1",
    ),
    (
        np.array(
            ["Apple", "Banana", "Cherry", "Date", "Fig", "Grapes"], dtype=np.str_
        ).reshape(2, 3),
        "s2",
    ),
    (
        np.array([""], dtype=np.str_).reshape(1, 1),
        "s3",
    ),
]


@pytest.mark.parametrize(
    "expected_array, var_name",
    params,
    ids=["simple-string-v7", "string-array-v7", "empty-string-v7"],
)
def test_string_load_v7(expected_array, var_name):
    file_path_v7 = os.path.join(os.path.dirname(__file__), "test_string_v7.mat")
    matdict = load_from_mat(file_path_v7, raw_data=False)

    assert var_name in matdict
    np.testing.assert_array_equal(matdict[var_name], expected_array)


@pytest.mark.parametrize(
    "expected_array, var_name",
    params,
    ids=["simple-string-v7.3", "string-array-v7.3", "empty-string-v7.3"],
)
def test_string_load_v73(expected_array, var_name):
    file_path_v73 = os.path.join(os.path.dirname(__file__), "test_string_v73.mat")
    matdict = load_from_mat(file_path_v73, raw_data=False)

    assert var_name in matdict
    np.testing.assert_array_equal(matdict[var_name], expected_array)



"""Utility functions for converting MATLAB datetime, duration, and calendarDuration"""

import warnings
from datetime import datetime
from zoneinfo import ZoneInfo, ZoneInfoNotFoundError

import numpy as np


def get_tz_offset(tz):
    """Get timezone offset in milliseconds (default UTC)"""
    try:
        tzinfo = ZoneInfo(tz)
        utc_offset = tzinfo.utcoffset(datetime.now())
        if utc_offset is not None:
            offset = int(utc_offset.total_seconds() * 1000)
        else:
            offset = 0
    except ZoneInfoNotFoundError as e:
        warnings.warn(
            f"mat_to_datetime: Could not get timezone offset for {tz}: {e}. Defaulting to UTC."
        )
        offset = 0
    return offset


def mat_to_datetime(props, **_kwargs):
    """Convert MATLAB datetime to Numpy datetime64 array"""

    data = props.get("data", np.array([]))
    if data.size == 0:
        return np.array([], dtype="datetime64[ms]")
    tz = props.get("tz", None)
    if tz is not None and tz.size > 0:
        offset = get_tz_offset(tz.item())
    else:
        offset = 0

    millis = data.real + data.imag * 1e3 + offset

    return millis.astype("datetime64[ms]")


def mat_to_duration(props, **_kwargs):
    """Convert MATLAB duration to Numpy timedelta64 array"""

    millis = props["millis"]
    if millis.size == 0:
        return np.array([], dtype="timedelta64[ms]")

    fmt = props.get("fmt", None)
    if fmt is None:
        return millis.astype("timedelta64[ms]")

    if fmt == "s":
        count = millis / 1000  # Seconds
        dur = count.astype("timedelta64[s]")
    elif fmt == "m":
        count = millis / (1000 * 60)  # Minutes
        dur = count.astype("timedelta64[m]")
    elif fmt == "h":
        count = millis / (1000 * 60 * 60)  # Hours
        dur = count.astype("timedelta64[h]")
    elif fmt == "d":
        count = millis / (1000 * 60 * 60 * 24)  # Days
        dur = count.astype("timedelta64[D]")
    elif fmt == "y":
        count = millis / (1000 * 60 * 60 * 24 * 365)  # Years
        dur = count.astype("timedelta64[Y]")
    else:
        count = millis
        dur = count.astype("timedelta64[ms]")
        # Default case

    return dur


def mat_to_calendarduration(props, **_kwargs):
    """Convert MATLAB calendarDuration to Dict of Python Timedeltas"""

    comps = props.get("components", None)
    if comps is None:
        return props

    comps[0, 0]["months"] = comps[0, 0]["months"].astype("timedelta64[M]")
    comps[0, 0]["days"] = comps[0, 0]["days"].astype("timedelta64[D]")
    comps[0, 0]["millis"] = comps[0, 0]["millis"].astype("timedelta64[ms]")

    return comps


def datetime_to_mat(arr):
    """Convert numpy.datetime64 array to MATLAB datetime format."""
    if not isinstance(arr, np.ndarray):
        raise TypeError(f"Expected numpy.ndarray, got {type(arr)}")
    if not np.issubdtype(arr.dtype, np.datetime64):
        raise TypeError(f"Expected numpy.datetime64 array, got {arr.dtype}")

    millis = arr.astype("datetime64[ms]").astype(np.float64)

    if "us" not in str(arr.dtype) and "ns" not in str(arr.dtype):
        data = arr.astype("datetime64[ms]").astype(np.float64)
    else:
        # For sub-ms precision
        millis = arr.astype("datetime64[ms]").astype(np.float64)
        us = arr.astype("datetime64[us]").astype(np.float64)
        frac_ms = (us % 1000) / 1000.0
        data = millis + 1j * frac_ms

    tz = np.empty((0, 0), dtype=np.str_)
    fmt = np.empty((0, 0), dtype=np.str_)
    prop_map = {"data": data, "tz": tz, "fmt": fmt}

    return prop_map


def duration_to_mat(arr):
    """Convert numpy timedelta64 array to MATLAB duration format."""
    if not isinstance(arr, np.ndarray):
        raise TypeError(f"Expected numpy.ndarray, got {type(arr)}")
    if not np.issubdtype(arr.dtype, np.timedelta64):
        raise TypeError(f"Expected numpy.timedelta64 array, got {arr.dtype}")

    unit, _ = np.datetime_data(arr.dtype)
    millis = arr.astype("timedelta64[ns]").astype(np.float64) / 1e6
    allowed_units = ("s", "m", "h", "D", "Y")
    if unit not in allowed_units:
        warnings.warn(
            f"duration_to_mat: MATLAB Duration arrays do not support timedelta64[{unit}]. Defaulting to 's'.",
            UserWarning,
        )
        unit = "s"

    unit = unit.lower()
    prop_map = {
        "millis": millis,
        "fmt": unit,
    }

    return prop_map


def calendarduration_to_mat(arr):
    """Convert numpy structured array with fields ['months', 'days', 'millis'] to MATLAB calendarDuration format."""
    if not isinstance(arr, np.ndarray):
        raise TypeError(f"Expected numpy.ndarray, got {type(arr)}")
    if arr.dtype.names is None or set(arr.dtype.names) != {"months", "days", "millis"}:
        raise TypeError(
            "Expected structured array with fields ['months', 'days', 'millis']"
        )
    if arr.size > 0 and arr.ndim == 1:
        arr = arr.reshape(1, -1)  # Ensure 2D Shape

    if arr.size > 0:
        arr[0, 0]["months"] = (
            arr[0, 0]["months"].astype("timedelta64[M]").astype("float64")
        )
        arr[0, 0]["days"] = arr[0, 0]["days"].astype("timedelta64[D]").astype("float64")
        arr[0, 0]["millis"] = (
            arr[0, 0]["millis"].astype("timedelta64[ms]").astype("float64")
        )

    fmt = "ymdt"

    prop_map = {
        "components": arr,
        "fmt": fmt,
    }

    return prop_map



"""Base class for MAT-file reading and writing"""

import struct
import sys
import time
from io import BytesIO

from matio.matio5 import read_matfile5, save_matfile5
from matio.matio7 import read_matfile7


def get_matfile_version(byte_data):
    """Reads subsystem MAT-file version and endianness from subsys byte stream"""

    maj_ind = int(byte_data[2] == b"I"[0])
    v_major = int(byte_data[maj_ind])
    v_minor = int(byte_data[1 - maj_ind])
    if v_major in (1, 2):
        return v_major, v_minor

    raise NotImplementedError(f"Unknown MAT-file version {v_major}.{v_minor}")


def load_from_mat(
    file_path,
    mdict=None,
    raw_data=False,
    add_table_attrs=False,
    *,
    spmatrix=True,
    **kwargs,
):
    """Loads variables from MAT-file
    Inputs
        1. file_path (str): Path to MAT-file
        2. mdict (dict): Dictionary to store loaded variables
        3. raw_data (bool): Whether to return raw data for objects
        4. add_table_attrs (bool): Add attributes to pandas DataFrame for MATLAB tables/timetables
        5. spmatrix (bool): Additional arguments for scipy.io.loadmat
        6. kwargs: Additional arguments for scipy.io.loadmat
    Returns:
        1. mdict (dict): Dictionary of loaded variables
    """

    with open(file_path, "rb") as f:
        f.seek(124)
        version_bytes = f.read(4)
        v_major, v_minor = get_matfile_version(version_bytes)

    if v_major == 1:
        matfile_dict = read_matfile5(
            file_path, raw_data, add_table_attrs, spmatrix, **kwargs
        )
    elif v_major == 2:
        matfile_dict = read_matfile7(
            file_path, raw_data, add_table_attrs, spmatrix, **kwargs
        )
    else:
        raise NotImplementedError(f"Unknown MAT-file version {v_major}.{v_minor}")

    # Update mdict if present
    if mdict is not None:
        mdict.update(matfile_dict)
    else:
        mdict = matfile_dict

    return mdict


def save_to_mat(
    file_path,
    mdict,
    version="v7",
    do_compression=True,
    global_vars=None,
    oned_as="row",
    use_strings=True,
):
    """Saves variables to MAT-file
    Inputs
        1. file_path (str): Path to MAT-file
        2. mdict (dict): Dictionary of variables to save
    Returns:
        None
    """

    mdict_copy = mdict.copy()
    # Copy as we are replacing objects in place with metadata arrays
    # This is because I'm using scipy.savemat to write data
    # Can be avoided with my own save implementation

    file_stream = BytesIO()
    write_file_header(file_stream, version)

    # FIXME: Should I include an argument for a saveobj return type?
    # Maybe as list of class names?

    if version == "v7":
        subsys_offset = save_matfile5(
            file_stream, mdict_copy, do_compression, global_vars, oned_as, use_strings
        )
    elif version == "v7.3":
        raise NotImplementedError("v7.3 MAT-file saving is not implemented")
    else:
        raise ValueError(f"Unknown MAT-file version '{version}' specified")

    if subsys_offset > 0:
        write_subsystem_offset(file_stream, subsys_offset)

    # FIXME: This might change when v7.3 gets implemented
    with open(file_path, "wb") as f:
        f.write(file_stream.getvalue())


def write_subsystem_offset(file_stream, offset=0):
    """Write 8 bytes of subsystem offset at byte 116"""

    file_stream.seek(116)
    file_stream.write(struct.pack("<Q", offset))
    file_stream.seek(0, 2)  # Seek to end


def write_version(file_stream, version):
    """Write version information"""

    if version == "v7":
        v_major, v_minor = 1, 0
    elif version == "v7.3":
        v_major, v_minor = 2, 0
    else:
        raise ValueError(f"Unknown version: {version}")

    is_little_endian = sys.byteorder == "little"

    if is_little_endian:
        file_stream.write(struct.pack("<BB", v_minor, v_major))
        file_stream.write(b"IM")
    else:
        file_stream.write(struct.pack(">BB", v_major, v_minor))
        file_stream.write(b"MI")


def write_file_header(file_stream, version):
    """Write MAT-file header"""

    current_time = time.strftime("%Y-%m-%d %H:%M:%S UTC", time.gmtime())
    description = (
        f"MATLAB 5.0 MAT-file Platform. "
        f"Created on: {current_time} by matio with scipy"
    )
    description_bytes = description.encode("ascii")[:116]  # Truncate if too long
    description_padded = description_bytes.ljust(116, b"\x20")

    file_stream.write(description_padded)
    write_subsystem_offset(file_stream)
    write_version(file_stream, version)
